@@ -19,23 +19,18 @@
 #include "Database.h"
 
 #include "core/AsyncTask.h"
-#include "core/Clock.h"
 #include "core/FileWatcher.h"
 #include "core/Group.h"
-#include "core/Merger.h"
-#include "core/Metadata.h"
 #include "format/KdbxXmlReader.h"
 #include "format/KeePass2Reader.h"
 #include "format/KeePass2Writer.h"
-#include "keys/FileKey.h"
-#include "keys/PasswordKey.h"
 
-#include <QFile>
 #include <QFileInfo>
+#include <QJsonObject>
+#include <QRegularExpression>
 #include <QSaveFile>
 #include <QTemporaryFile>
 #include <QTimer>
-#include <QXmlStreamReader>
 
 QHash<QUuid, QPointer<Database>> Database::s_uuidMap;
 
@@ -44,24 +39,39 @@ Database::Database()
     , m_data()
     , m_rootGroup(nullptr)
     , m_fileWatcher(new FileWatcher(this))
-    , m_emitModified(false)
     , m_uuid(QUuid::createUuid())
 {
+    // setup modified timer
+    m_modifiedTimer.setSingleShot(true);
+    connect(this, &Database::emitModifiedChanged, this, [this](bool value) {
+        if (!value) {
+            stopModifiedTimer();
+        }
+    });
+    connect(&m_modifiedTimer, &QTimer::timeout, this, &Database::emitModified);
+
+    // other signals
+    connect(m_metadata, &Metadata::modified, this, &Database::markAsModified);
+    connect(this, &Database::databaseOpened, this, [this]() {
+        updateCommonUsernames();
+        updateTagList();
+    });
+    connect(this, &Database::modified, this, [this] { updateTagList(); });
+    connect(this, &Database::databaseSaved, this, [this]() { updateCommonUsernames(); });
+    connect(m_fileWatcher, &FileWatcher::fileChanged, this, &Database::databaseFileChanged);
+
+    // static uuid map
+    s_uuidMap.insert(m_uuid, this);
+
+    // block modified signal and set root group
+    setEmitModified(false);
+
     setRootGroup(new Group());
     rootGroup()->setUuid(QUuid::createUuid());
     rootGroup()->setName(tr("Passwords", "Root group name"));
-    m_modifiedTimer.setSingleShot(true);
-
-    s_uuidMap.insert(m_uuid, this);
-
-    connect(m_metadata, SIGNAL(metadataModified()), SLOT(markAsModified()));
-    connect(&m_modifiedTimer, SIGNAL(timeout()), SIGNAL(databaseModified()));
-    connect(this, SIGNAL(databaseOpened()), SLOT(updateCommonUsernames()));
-    connect(this, SIGNAL(databaseSaved()), SLOT(updateCommonUsernames()));
-    connect(m_fileWatcher, &FileWatcher::fileChanged, this, &Database::databaseFileChanged);
 
     m_modified = false;
-    m_emitModified = true;
+    setEmitModified(true);
 }
 
 Database::Database(const QString& filePath)
@@ -86,17 +96,16 @@ QUuid Database::uuid() const
  * read-write mode and fall back to read-only if that is not possible.
  *
  * @param key composite key for unlocking the database
- * @param readOnly open in read-only mode
  * @param error error message in case of failure
  * @return true on success
  */
-bool Database::open(QSharedPointer<const CompositeKey> key, QString* error, bool readOnly)
+bool Database::open(QSharedPointer<const CompositeKey> key, QString* error)
 {
     Q_ASSERT(!m_data.filePath.isEmpty());
     if (m_data.filePath.isEmpty()) {
         return false;
     }
-    return open(m_data.filePath, std::move(key), error, readOnly);
+    return open(m_data.filePath, std::move(key), error);
 }
 
 /**
@@ -106,11 +115,10 @@ bool Database::open(QSharedPointer<const CompositeKey> key, QString* error, bool
  *
  * @param filePath path to the file
  * @param key composite key for unlocking the database
- * @param readOnly open in read-only mode
  * @param error error message in case of failure
  * @return true on success
  */
-bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey> key, QString* error, bool readOnly)
+bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey> key, QString* error)
 {
     QFile dbFile(filePath);
     if (!dbFile.exists()) {
@@ -144,7 +152,6 @@ bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey>
         return false;
     }
 
-    setReadOnly(readOnly);
     setFilePath(filePath);
     dbFile.close();
 
@@ -157,6 +164,27 @@ bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey>
     return true;
 }
 
+/**
+ * KDBX format version.
+ */
+quint32 Database::formatVersion() const
+{
+    return m_data.formatVersion;
+}
+
+void Database::setFormatVersion(quint32 version)
+{
+    m_data.formatVersion = version;
+}
+
+/**
+ * Whether the KDBX minor version is greater than the newest supported.
+ */
+bool Database::hasMinorVersionMismatch() const
+{
+    return m_data.formatVersion > KeePass2::FILE_VERSION_MAX;
+}
+
 bool Database::isSaving()
 {
     bool locked = m_saveMutex.tryLock();
@@ -172,10 +200,10 @@ bool Database::isSaving()
  *
  * @param error error message in case of failure
  * @param atomic Use atomic file transactions
- * @param backup Backup the existing database file, if exists
+ * @param backupFilePath Absolute file path to write the backup file to. Pass an empty QString to disable backup.
  * @return true on success
  */
-bool Database::save(QString* error, bool atomic, bool backup)
+bool Database::save(SaveAction action, const QString& backupFilePath, QString* error)
 {
     Q_ASSERT(!m_data.filePath.isEmpty());
     if (m_data.filePath.isEmpty()) {
@@ -185,7 +213,7 @@ bool Database::save(QString* error, bool atomic, bool backup)
         return false;
     }
 
-    return saveAs(m_data.filePath, error, atomic, backup);
+    return saveAs(m_data.filePath, action, backupFilePath, error);
 }
 
 /**
@@ -203,10 +231,11 @@ bool Database::save(QString* error, bool atomic, bool backup)
  * @param filePath Absolute path of the file to save
  * @param error error message in case of failure
  * @param atomic Use atomic file transactions
- * @param backup Backup the existing database file, if exists
+ * @param backupFilePath Absolute path to the location where the backup should be stored. Passing an empty string
+ * disables backup.
  * @return true on success
  */
-bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool backup)
+bool Database::saveAs(const QString& filePath, SaveAction action, const QString& backupFilePath, QString* error)
 {
     // Disallow overlapping save operations
     if (isSaving()) {
@@ -228,15 +257,6 @@ bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool
     QMutexLocker locker(&m_saveMutex);
 
     if (filePath == m_data.filePath) {
-        // Disallow saving to the same file if read-only
-        if (m_data.isReadOnly) {
-            Q_ASSERT_X(false, "Database::saveAs", "Could not save, database file is read-only.");
-            if (error) {
-                *error = tr("Could not save, database file is read-only.");
-            }
-            return false;
-        }
-
         // Fail-safe check to make sure we don't overwrite underlying file changes
         // that have not yet triggered a file reload/merge operation.
         if (!m_fileWatcher->hasSameFileChecksum()) {
@@ -248,13 +268,12 @@ bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool
     }
 
     // Clear read-only flag
-    setReadOnly(false);
     m_fileWatcher->stop();
 
     QFileInfo fileInfo(filePath);
     auto realFilePath = fileInfo.exists() ? fileInfo.canonicalFilePath() : fileInfo.absoluteFilePath();
     bool isNewFile = !QFile::exists(realFilePath);
-    bool ok = AsyncTask::runAndWaitForFuture([&] { return performSave(realFilePath, error, atomic, backup); });
+    bool ok = AsyncTask::runAndWaitForFuture([&] { return performSave(realFilePath, action, backupFilePath, error); });
     if (ok) {
         markAsClean();
         setFilePath(filePath);
@@ -270,14 +289,19 @@ bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool
     return ok;
 }
 
-bool Database::performSave(const QString& filePath, QString* error, bool atomic, bool backup)
+bool Database::performSave(const QString& filePath, SaveAction action, const QString& backupFilePath, QString* error)
 {
+    if (!backupFilePath.isNull()) {
+        backupDatabase(filePath, backupFilePath);
+    }
+
 #if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
     QFileInfo info(filePath);
     auto createTime = info.exists() ? info.birthTime() : QDateTime::currentDateTime();
 #endif
 
-    if (atomic) {
+    switch (action) {
+    case Atomic: {
         QSaveFile saveFile(filePath);
         if (saveFile.open(QIODevice::WriteOnly)) {
             // write the database to the file
@@ -285,12 +309,8 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
                 return false;
             }
 
-            if (backup) {
-                backupDatabase(filePath);
-            }
-
 #if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
-            // Retain orginal creation time
+            // Retain original creation time
             saveFile.setFileTime(createTime, QFile::FileBirthTime);
 #endif
 
@@ -303,20 +323,17 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
         if (error) {
             *error = saveFile.errorString();
         }
-    } else {
+        break;
+    }
+    case TempFile: {
         QTemporaryFile tempFile;
         if (tempFile.open()) {
             // write the database to the file
             if (!writeDatabase(&tempFile, error)) {
                 return false;
             }
-
             tempFile.close(); // flush to disk
 
-            if (backup) {
-                backupDatabase(filePath);
-            }
-
             // Delete the original db and move the temp file in place
             auto perms = QFile::permissions(filePath);
             QFile::remove(filePath);
@@ -333,7 +350,7 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
                 tempFile.setFileTime(createTime, QFile::FileBirthTime);
 #endif
                 return true;
-            } else if (!backup || !restoreDatabase(filePath)) {
+            } else if (backupFilePath.isEmpty() || !restoreDatabase(filePath, backupFilePath)) {
                 // Failed to copy new database in place, and
                 // failed to restore from backup or backups disabled
                 tempFile.setAutoRemove(false);
@@ -347,6 +364,23 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
         if (error) {
             *error = tempFile.errorString();
         }
+        break;
+    }
+    case DirectWrite: {
+        // Open the original database file for direct-write
+        QFile dbFile(filePath);
+        if (dbFile.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
+            if (!writeDatabase(&dbFile, error)) {
+                return false;
+            }
+            dbFile.close();
+            return true;
+        }
+        if (error) {
+            *error = dbFile.errorString();
+        }
+        break;
+    }
     }
 
     // Saving failed
@@ -355,17 +389,9 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
 
 bool Database::writeDatabase(QIODevice* device, QString* error)
 {
-    Q_ASSERT(!m_data.isReadOnly);
-    if (m_data.isReadOnly) {
-        if (error) {
-            *error = tr("File cannot be written as it is opened in read-only mode.");
-        }
-        return false;
-    }
-
     PasswordKey oldTransformedKey;
     if (m_data.key->isEmpty()) {
-        oldTransformedKey.setHash(m_data.transformedDatabaseKey->rawKey());
+        oldTransformedKey.setRawKey(m_data.transformedDatabaseKey->rawKey());
     }
 
     KeePass2Writer writer;
@@ -385,7 +411,7 @@ bool Database::writeDatabase(QIODevice* device, QString* error)
     Q_ASSERT(newKey != oldTransformedKey.rawKey());
     if (newKey.isEmpty() || newKey == oldTransformedKey.rawKey()) {
         if (error) {
-            *error = tr("Key not transformed. This is a bug, please report it to the developers!");
+            *error = tr("Key not transformed. This is a bug, please report it to the developers.");
         }
         return false;
     }
@@ -445,7 +471,6 @@ void Database::releaseData()
 
     setEmitModified(false);
     m_modified = false;
-    m_modifiedTimer.stop();
 
     s_uuidMap.remove(m_uuid);
     m_uuid = QUuid();
@@ -453,32 +478,39 @@ void Database::releaseData()
     m_data.clear();
     m_metadata->clear();
 
+    auto oldGroup = rootGroup();
     setRootGroup(new Group());
+    // explicitly delete old group, otherwise it is only deleted when the database object is destructed
+    delete oldGroup;
 
     m_fileWatcher->stop();
 
     m_deletedObjects.clear();
     m_commonUsernames.clear();
+    m_tagList.clear();
 }
 
 /**
- * Remove the old backup and replace it with a new one
- * backups are named <filename>.old.<extension>
+ * Remove the old backup and replace it with a new one. Backup name is taken from destinationFilePath.
+ * Non-existing parent directories will be created automatically.
  *
  * @param filePath Path to the file to backup
+ * @param destinationFilePath Path to the backup destination file
  * @return true on success
  */
-bool Database::backupDatabase(const QString& filePath)
+bool Database::backupDatabase(const QString& filePath, const QString& destinationFilePath)
 {
-    static auto re = QRegularExpression("(\\.[^.]+)$");
-
-    auto match = re.match(filePath);
-    auto backupFilePath = filePath;
+    // Ensure that the path to write to actually exists
+    auto parentDirectory = QFileInfo(destinationFilePath).absoluteDir();
+    if (!parentDirectory.exists()) {
+        if (!QDir().mkpath(parentDirectory.absolutePath())) {
+            return false;
+        }
+    }
     auto perms = QFile::permissions(filePath);
-    backupFilePath = backupFilePath.replace(re, "") + ".old" + match.captured(1);
-    QFile::remove(backupFilePath);
-    bool res = QFile::copy(filePath, backupFilePath);
-    QFile::setPermissions(backupFilePath, perms);
+    QFile::remove(destinationFilePath);
+    bool res = QFile::copy(filePath, destinationFilePath);
+    QFile::setPermissions(destinationFilePath, perms);
     return res;
 }
 
@@ -490,33 +522,19 @@ bool Database::backupDatabase(const QString& filePath)
  * @param filePath Path to the file to restore
  * @return true on success
  */
-bool Database::restoreDatabase(const QString& filePath)
+bool Database::restoreDatabase(const QString& filePath, const QString& fromBackupFilePath)
 {
-    static auto re = QRegularExpression("^(.*?)(\\.[^.]+)?$");
-
-    auto match = re.match(filePath);
     auto perms = QFile::permissions(filePath);
-    auto backupFilePath = match.captured(1) + ".old" + match.captured(2);
     // Only try to restore if the backup file actually exists
-    if (QFile::exists(backupFilePath)) {
+    if (QFile::exists(fromBackupFilePath)) {
         QFile::remove(filePath);
-        if (QFile::copy(backupFilePath, filePath)) {
+        if (QFile::copy(fromBackupFilePath, filePath)) {
             return QFile::setPermissions(filePath, perms);
         }
     }
     return false;
 }
 
-bool Database::isReadOnly() const
-{
-    return m_data.isReadOnly;
-}
-
-void Database::setReadOnly(bool readOnly)
-{
-    m_data.isReadOnly = readOnly;
-}
-
 /**
  * Returns true if the database key exists, has subkeys, and the
  * root group exists
@@ -656,17 +674,46 @@ void Database::addDeletedObject(const QUuid& uuid)
     addDeletedObject(delObj);
 }
 
-QList<QString> Database::commonUsernames()
+const QStringList& Database::commonUsernames() const
 {
     return m_commonUsernames;
 }
 
+const QStringList& Database::tagList() const
+{
+    return m_tagList;
+}
+
 void Database::updateCommonUsernames(int topN)
 {
     m_commonUsernames.clear();
     m_commonUsernames.append(rootGroup()->usernamesRecursive(topN));
 }
 
+void Database::updateTagList()
+{
+    m_tagList.clear();
+    if (!m_rootGroup) {
+        emit tagListUpdated();
+        return;
+    }
+
+    // Search groups recursively looking for tags
+    // Use a set to prevent adding duplicates
+    QSet<QString> tagSet;
+    for (const auto group : m_rootGroup->groupsRecursive(true)) {
+        for (const auto entry : group->entries()) {
+            for (auto tag : entry->tagList()) {
+                tagSet.insert(tag);
+            }
+        }
+    }
+
+    m_tagList = tagSet.toList();
+    m_tagList.sort();
+    emit tagListUpdated();
+}
+
 const QUuid& Database::cipher() const
 {
     return m_data.cipher;
@@ -691,11 +738,11 @@ bool Database::challengeMasterSeed(const QByteArray& masterSeed)
 {
     m_keyError.clear();
     if (m_data.key) {
-        m_data.masterSeed->setHash(masterSeed);
+        m_data.masterSeed->setRawKey(masterSeed);
         QByteArray response;
         bool ok = m_data.key->challenge(masterSeed, response, &m_keyError);
         if (ok && !response.isEmpty()) {
-            m_data.challengeResponseKey->setHash(response);
+            m_data.challengeResponseKey->setRawKey(response);
         } else if (ok && response.isEmpty()) {
             // no CR key present, make sure buffer is empty
             m_data.challengeResponseKey.reset(new PasswordKey);
@@ -733,7 +780,6 @@ bool Database::setKey(const QSharedPointer<const CompositeKey>& key,
                       bool updateTransformSalt,
                       bool transformKey)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     m_keyError.clear();
 
     if (!key) {
@@ -749,7 +795,7 @@ bool Database::setKey(const QSharedPointer<const CompositeKey>& key,
 
     PasswordKey oldTransformedDatabaseKey;
     if (m_data.key && !m_data.key->isEmpty()) {
-        oldTransformedDatabaseKey.setHash(m_data.transformedDatabaseKey->rawKey());
+        oldTransformedDatabaseKey.setRawKey(m_data.transformedDatabaseKey->rawKey());
     }
 
     QByteArray transformedDatabaseKey;
@@ -762,7 +808,7 @@ bool Database::setKey(const QSharedPointer<const CompositeKey>& key,
 
     m_data.key = key;
     if (!transformedDatabaseKey.isEmpty()) {
-        m_data.transformedDatabaseKey->setHash(transformedDatabaseKey);
+        m_data.transformedDatabaseKey->setRawKey(transformedDatabaseKey);
     }
     if (updateChangedTime) {
         m_metadata->setDatabaseKeyChanged(Clock::currentDateTimeUtc());
@@ -792,14 +838,11 @@ const QVariantMap& Database::publicCustomData() const
 
 void Database::setPublicCustomData(const QVariantMap& customData)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     m_data.publicCustomData = customData;
 }
 
 void Database::createRecycleBin()
 {
-    Q_ASSERT(!m_data.isReadOnly);
-
     auto recycleBin = new Group();
     recycleBin->setUuid(QUuid::createUuid());
     recycleBin->setParent(rootGroup());
@@ -813,7 +856,6 @@ void Database::createRecycleBin()
 
 void Database::recycleEntry(Entry* entry)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     if (m_metadata->recycleBinEnabled()) {
         if (!m_metadata->recycleBin()) {
             createRecycleBin();
@@ -826,7 +868,6 @@ void Database::recycleEntry(Entry* entry)
 
 void Database::recycleGroup(Group* group)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     if (m_metadata->recycleBinEnabled()) {
         if (!m_metadata->recycleBin()) {
             createRecycleBin();
@@ -839,7 +880,6 @@ void Database::recycleGroup(Group* group)
 
 void Database::emptyRecycleBin()
 {
-    Q_ASSERT(!m_data.isReadOnly);
     if (m_metadata->recycleBinEnabled() && m_metadata->recycleBin()) {
         // destroying direct entries of the recycle bin
         QList<Entry*> subEntries = m_metadata->recycleBin()->entries();
@@ -854,15 +894,6 @@ void Database::emptyRecycleBin()
     }
 }
 
-void Database::setEmitModified(bool value)
-{
-    if (m_emitModified && !value) {
-        m_modifiedTimer.stop();
-    }
-
-    m_emitModified = value;
-}
-
 bool Database::isModified() const
 {
     return m_modified;
@@ -876,9 +907,9 @@ bool Database::hasNonDataChanges() const
 void Database::markAsModified()
 {
     m_modified = true;
-    if (m_emitModified && !m_modifiedTimer.isActive()) {
+    if (modifiedSignalEnabled() && !m_modifiedTimer.isActive()) {
         // Small time delay prevents numerous consecutive saves due to repeated signals
-        m_modifiedTimer.start(150);
+        startModifiedTimer();
     }
 }
 
@@ -886,7 +917,7 @@ void Database::markAsClean()
 {
     bool emitSignal = m_modified;
     m_modified = false;
-    m_modifiedTimer.stop();
+    stopModifiedTimer();
     m_hasNonDataChange = false;
     if (emitSignal) {
         emit databaseSaved();
@@ -919,14 +950,12 @@ QSharedPointer<Kdf> Database::kdf() const
 
 void Database::setKdf(QSharedPointer<Kdf> kdf)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     m_data.kdf = std::move(kdf);
+    setFormatVersion(KeePass2Writer::kdbxVersionRequired(this, true, m_data.kdf.isNull()));
 }
 
 bool Database::changeKdf(const QSharedPointer<Kdf>& kdf)
 {
-    Q_ASSERT(!m_data.isReadOnly);
-
     kdf->randomizeSeed();
     QByteArray transformedDatabaseKey;
     if (!m_data.key) {
@@ -937,8 +966,20 @@ bool Database::changeKdf(const QSharedPointer<Kdf>& kdf)
     }
 
     setKdf(kdf);
-    m_data.transformedDatabaseKey->setHash(transformedDatabaseKey);
+    m_data.transformedDatabaseKey->setRawKey(transformedDatabaseKey);
     markAsModified();
 
     return true;
 }
+
+// Prevent warning about QTimer not allowed to be started/stopped from other thread
+void Database::startModifiedTimer()
+{
+    QMetaObject::invokeMethod(&m_modifiedTimer, "start", Q_ARG(int, 150));
+}
+
+// Prevent warning about QTimer not allowed to be started/stopped from other thread
+void Database::stopModifiedTimer()
+{
+    QMetaObject::invokeMethod(&m_modifiedTimer, "stop");
+}
