@@ -1,6 +1,6 @@
 /*
  *  Copyright (C) 2010 Felix Geyer <debfx@fobos.de>
- *  Copyright (C) 2017 KeePassXC Team <team@keepassxc.org>
+ *  Copyright (C) 2020 KeePassXC Team <team@keepassxc.org>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,34 +22,27 @@
 #include <QCloseEvent>
 #include <QDesktopServices>
 #include <QFileInfo>
+#include <QList>
 #include <QMimeData>
 #include <QShortcut>
+#include <QStatusBar>
 #include <QTimer>
+#include <QToolButton>
 #include <QWindow>
 
 #include "config-keepassx.h"
 
+#include "Application.h"
+#include "Clipboard.h"
 #include "autotype/AutoType.h"
-#include "core/Config.h"
 #include "core/InactivityTimer.h"
-#include "core/Metadata.h"
 #include "core/Resources.h"
 #include "core/Tools.h"
 #include "gui/AboutDialog.h"
-#include "gui/DatabaseWidget.h"
+#include "gui/Icons.h"
 #include "gui/MessageBox.h"
 #include "gui/SearchWidget.h"
 #include "gui/osutils/OSUtils.h"
-#include "keys/CompositeKey.h"
-#include "keys/FileKey.h"
-#include "keys/PasswordKey.h"
-
-#ifdef Q_OS_MACOS
-#include "gui/osutils/macutils/MacUtils.h"
-#ifdef WITH_XC_TOUCHID
-#include "touchid/TouchID.h"
-#endif
-#endif
 
 #ifdef WITH_XC_UPDATECHECK
 #include "gui/UpdateCheckDialog.h"
@@ -79,9 +72,7 @@
 #endif
 
 #if defined(Q_OS_UNIX) && !defined(Q_OS_MACOS) && !defined(QT_NO_DBUS)
-#include "gui/MainWindowAdaptor.h"
-#include <QList>
-#include <QtDBus/QtDBus>
+#include "mainwindowadaptor.h"
 #endif
 
 const QString MainWindow::BaseWindowTitle = "KeePassXC";
@@ -132,13 +123,13 @@ MainWindow::MainWindow()
     m_countDefaultAttributes = m_ui->menuEntryCopyAttribute->actions().size();
 
     m_entryContextMenu = new QMenu(this);
+    m_entryContextMenu->setSeparatorsCollapsible(true);
     m_entryContextMenu->addAction(m_ui->actionEntryCopyUsername);
     m_entryContextMenu->addAction(m_ui->actionEntryCopyPassword);
     m_entryContextMenu->addAction(m_ui->menuEntryCopyAttribute->menuAction());
     m_entryContextMenu->addAction(m_ui->menuEntryTotp->menuAction());
     m_entryContextMenu->addSeparator();
     m_entryContextMenu->addAction(m_ui->actionEntryAutoType);
-    m_entryContextMenu->addAction(m_ui->menuEntryAutoTypeWithSequence->menuAction());
     m_entryContextMenu->addSeparator();
     m_entryContextMenu->addAction(m_ui->actionEntryEdit);
     m_entryContextMenu->addAction(m_ui->actionEntryClone);
@@ -150,20 +141,45 @@ MainWindow::MainWindow()
     m_entryContextMenu->addSeparator();
     m_entryContextMenu->addAction(m_ui->actionEntryOpenUrl);
     m_entryContextMenu->addAction(m_ui->actionEntryDownloadIcon);
+    m_entryContextMenu->addSeparator();
+    m_entryContextMenu->addAction(m_ui->actionEntryRestore);
 
     m_entryNewContextMenu = new QMenu(this);
     m_entryNewContextMenu->addAction(m_ui->actionEntryNew);
 
+    // Build Entry Level Auto-Type menu
+    auto autotypeMenu = new QMenu({}, this);
+    autotypeMenu->addAction(m_ui->actionEntryAutoTypeSequence);
+    autotypeMenu->addSeparator();
+    autotypeMenu->addAction(m_ui->actionEntryAutoTypeUsername);
+    autotypeMenu->addAction(m_ui->actionEntryAutoTypeUsernameEnter);
+    autotypeMenu->addAction(m_ui->actionEntryAutoTypePassword);
+    autotypeMenu->addAction(m_ui->actionEntryAutoTypePasswordEnter);
+    autotypeMenu->addAction(m_ui->actionEntryAutoTypeTOTP);
+    m_ui->actionEntryAutoType->setMenu(autotypeMenu);
+    auto autoTypeButton = qobject_cast<QToolButton*>(m_ui->toolBar->widgetForAction(m_ui->actionEntryAutoType));
+    if (autoTypeButton) {
+        autoTypeButton->setPopupMode(QToolButton::MenuButtonPopup);
+    }
+
+    auto databaseLockMenu = new QMenu({}, this);
+    databaseLockMenu->addAction(m_ui->actionLockAllDatabases);
+    m_ui->actionLockDatabaseToolbar->setMenu(databaseLockMenu);
+    auto databaseLockButton =
+        qobject_cast<QToolButton*>(m_ui->toolBar->widgetForAction(m_ui->actionLockDatabaseToolbar));
+    if (databaseLockButton) {
+        databaseLockButton->setPopupMode(QToolButton::MenuButtonPopup);
+    }
+
     restoreGeometry(config()->get(Config::GUI_MainWindowGeometry).toByteArray());
     restoreState(config()->get(Config::GUI_MainWindowState).toByteArray());
+
+    connect(m_ui->tabWidget, &DatabaseTabWidget::databaseLocked, this, &MainWindow::databaseLocked);
+    connect(m_ui->tabWidget, &DatabaseTabWidget::databaseUnlocked, this, &MainWindow::databaseUnlocked);
+    connect(m_ui->tabWidget, &DatabaseTabWidget::activeDatabaseChanged, this, &MainWindow::activeDatabaseChanged);
+
 #ifdef WITH_XC_BROWSER
     m_ui->settingsWidget->addSettingsPage(new BrowserSettingsPage());
-    connect(m_ui->tabWidget, &DatabaseTabWidget::databaseLocked, browserService(), &BrowserService::databaseLocked);
-    connect(m_ui->tabWidget, &DatabaseTabWidget::databaseUnlocked, browserService(), &BrowserService::databaseUnlocked);
-    connect(m_ui->tabWidget,
-            &DatabaseTabWidget::activateDatabaseChanged,
-            browserService(),
-            &BrowserService::activeDatabaseChanged);
     connect(
         browserService(), &BrowserService::requestUnlock, m_ui->tabWidget, &DatabaseTabWidget::performBrowserUnlock);
 #endif
@@ -177,8 +193,8 @@ MainWindow::MainWindow()
     m_entryContextMenu->addAction(m_ui->actionEntryAddToAgent);
     m_entryContextMenu->addAction(m_ui->actionEntryRemoveFromAgent);
 
-    m_ui->actionEntryAddToAgent->setIcon(resources()->icon("utilities-terminal"));
-    m_ui->actionEntryRemoveFromAgent->setIcon(resources()->icon("utilities-terminal"));
+    m_ui->actionEntryAddToAgent->setIcon(icons()->icon("utilities-terminal"));
+    m_ui->actionEntryRemoveFromAgent->setIcon(icons()->icon("utilities-terminal"));
 #endif
 
     m_ui->actionEntryAddToAgent->setVisible(false);
@@ -208,7 +224,7 @@ MainWindow::MainWindow()
     connect(YubiKey::instance(), SIGNAL(challengeCompleted()), SLOT(hideYubiKeyPopup()), Qt::QueuedConnection);
 #endif
 
-    setWindowIcon(resources()->applicationIcon());
+    setWindowIcon(icons()->applicationIcon());
     m_ui->globalMessageWidget->hideMessage();
     connect(m_ui->globalMessageWidget, &MessageWidget::linkActivated, &MessageWidget::openHttpUrl);
 
@@ -233,19 +249,10 @@ MainWindow::MainWindow()
     m_ui->toolbarSeparator->setVisible(false);
     m_showToolbarSeparator = config()->get(Config::GUI_ApplicationTheme).toString() != "classic";
 
-    bool isAutoTypeAvailable = autoType()->isAvailable();
-    m_ui->actionEntryAutoType->setVisible(isAutoTypeAvailable);
-    m_ui->actionEntryAutoTypeUsername->setVisible(isAutoTypeAvailable);
-    m_ui->actionEntryAutoTypeUsernameEnter->setVisible(isAutoTypeAvailable);
-    m_ui->actionEntryAutoTypePassword->setVisible(isAutoTypeAvailable);
-    m_ui->actionEntryAutoTypePasswordEnter->setVisible(isAutoTypeAvailable);
+    m_ui->actionEntryAutoType->setVisible(autoType()->isAvailable());
 
     m_inactivityTimer = new InactivityTimer(this);
     connect(m_inactivityTimer, SIGNAL(inactivityDetected()), this, SLOT(lockDatabasesAfterInactivity()));
-#ifdef WITH_XC_TOUCHID
-    m_touchIDinactivityTimer = new InactivityTimer(this);
-    connect(m_touchIDinactivityTimer, SIGNAL(inactivityDetected()), this, SLOT(forgetTouchIDAfterInactivity()));
-#endif
     applySettingsChanges();
 
     m_ui->actionDatabaseNew->setShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_N);
@@ -253,7 +260,8 @@ MainWindow::MainWindow()
     setShortcut(m_ui->actionDatabaseSave, QKeySequence::Save, Qt::CTRL + Qt::Key_S);
     setShortcut(m_ui->actionDatabaseSaveAs, QKeySequence::SaveAs, Qt::CTRL + Qt::SHIFT + Qt::Key_S);
     setShortcut(m_ui->actionDatabaseClose, QKeySequence::Close, Qt::CTRL + Qt::Key_W);
-    m_ui->actionLockDatabases->setShortcut(Qt::CTRL + Qt::Key_L);
+    m_ui->actionLockDatabase->setShortcut(Qt::CTRL + Qt::Key_L);
+    m_ui->actionLockAllDatabases->setShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_L);
     setShortcut(m_ui->actionQuit, QKeySequence::Quit, Qt::CTRL + Qt::Key_Q);
     setShortcut(m_ui->actionEntryNew, QKeySequence::New, Qt::CTRL + Qt::Key_N);
     m_ui->actionEntryEdit->setShortcut(Qt::CTRL + Qt::Key_E);
@@ -267,9 +275,10 @@ MainWindow::MainWindow()
     m_ui->actionEntryMoveDown->setShortcut(Qt::CTRL + Qt::ALT + Qt::Key_Down);
     m_ui->actionEntryCopyUsername->setShortcut(Qt::CTRL + Qt::Key_B);
     m_ui->actionEntryCopyPassword->setShortcut(Qt::CTRL + Qt::Key_C);
-    m_ui->actionEntryAutoType->setShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_V);
+    m_ui->actionEntryAutoTypeSequence->setShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_V);
     m_ui->actionEntryOpenUrl->setShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_U);
     m_ui->actionEntryCopyURL->setShortcut(Qt::CTRL + Qt::Key_U);
+    m_ui->actionEntryRestore->setShortcut(Qt::CTRL + Qt::Key_R);
 
     // Prevent conflicts with global Mac shortcuts (force Control on all platforms)
 #ifdef Q_OS_MAC
@@ -286,6 +295,7 @@ MainWindow::MainWindow()
     m_ui->actionEntryNew->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryEdit->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryDelete->setShortcutVisibleInContextMenu(true);
+    m_ui->actionEntryRestore->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryClone->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryTotp->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryDownloadIcon->setShortcutVisibleInContextMenu(true);
@@ -294,7 +304,7 @@ MainWindow::MainWindow()
     m_ui->actionEntryMoveDown->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryCopyUsername->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryCopyPassword->setShortcutVisibleInContextMenu(true);
-    m_ui->actionEntryAutoType->setShortcutVisibleInContextMenu(true);
+    m_ui->actionEntryAutoTypeSequence->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryOpenUrl->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryCopyURL->setShortcutVisibleInContextMenu(true);
     m_ui->actionEntryAddToAgent->setShortcutVisibleInContextMenu(true);
@@ -314,9 +324,14 @@ MainWindow::MainWindow()
     new QShortcut(Qt::CTRL + Qt::Key_M, this, SLOT(minimizeOrHide()));
     new QShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_M, this, SLOT(hideWindow()));
     // Control database tabs
-    new QShortcut(Qt::CTRL + Qt::Key_Tab, this, SLOT(selectNextDatabaseTab()));
+    // Ctrl+Tab is broken on Mac, so use Alt (i.e. the Option key) - https://bugreports.qt.io/browse/QTBUG-8596
+    auto dbTabModifier2 = Qt::CTRL;
+#ifdef Q_OS_MACOS
+    dbTabModifier2 = Qt::ALT;
+#endif
+    new QShortcut(dbTabModifier2 + Qt::Key_Tab, this, SLOT(selectNextDatabaseTab()));
     new QShortcut(Qt::CTRL + Qt::Key_PageDown, this, SLOT(selectNextDatabaseTab()));
-    new QShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_Tab, this, SLOT(selectPreviousDatabaseTab()));
+    new QShortcut(dbTabModifier2 + Qt::SHIFT + Qt::Key_Tab, this, SLOT(selectPreviousDatabaseTab()));
     new QShortcut(Qt::CTRL + Qt::Key_PageUp, this, SLOT(selectPreviousDatabaseTab()));
 
     // Tab selection by number, Windows uses Ctrl, macOS uses Command,
@@ -344,59 +359,64 @@ MainWindow::MainWindow()
     shortcut = new QShortcut(dbTabModifier + Qt::Key_9, this);
     connect(shortcut, &QShortcut::activated, [this]() { selectDatabaseTab(m_ui->tabWidget->count() - 1); });
 
-    m_ui->actionDatabaseNew->setIcon(resources()->icon("document-new"));
-    m_ui->actionDatabaseOpen->setIcon(resources()->icon("document-open"));
-    m_ui->menuRecentDatabases->setIcon(resources()->icon("document-open-recent"));
-    m_ui->actionDatabaseSave->setIcon(resources()->icon("document-save"));
-    m_ui->actionDatabaseSaveAs->setIcon(resources()->icon("document-save-as"));
-    m_ui->actionDatabaseSaveBackup->setIcon(resources()->icon("document-save-copy"));
-    m_ui->actionDatabaseClose->setIcon(resources()->icon("document-close"));
-    m_ui->actionReports->setIcon(resources()->icon("reports"));
-    m_ui->actionDatabaseSettings->setIcon(resources()->icon("document-edit"));
-    m_ui->actionDatabaseSecurity->setIcon(resources()->icon("database-change-key"));
-    m_ui->actionLockDatabases->setIcon(resources()->icon("database-lock"));
-    m_ui->actionQuit->setIcon(resources()->icon("application-exit"));
-    m_ui->actionDatabaseMerge->setIcon(resources()->icon("database-merge"));
-    m_ui->menuImport->setIcon(resources()->icon("document-import"));
-    m_ui->menuExport->setIcon(resources()->icon("document-export"));
+    m_ui->actionDatabaseNew->setIcon(icons()->icon("document-new"));
+    m_ui->actionDatabaseOpen->setIcon(icons()->icon("document-open"));
+    m_ui->menuRecentDatabases->setIcon(icons()->icon("document-open-recent"));
+    m_ui->actionDatabaseSave->setIcon(icons()->icon("document-save"));
+    m_ui->actionDatabaseSaveAs->setIcon(icons()->icon("document-save-as"));
+    m_ui->actionDatabaseSaveBackup->setIcon(icons()->icon("document-save-copy"));
+    m_ui->actionDatabaseClose->setIcon(icons()->icon("document-close"));
+    m_ui->actionReports->setIcon(icons()->icon("reports"));
+    m_ui->actionDatabaseSettings->setIcon(icons()->icon("document-edit"));
+    m_ui->actionDatabaseSecurity->setIcon(icons()->icon("database-change-key"));
+    m_ui->actionLockDatabase->setIcon(icons()->icon("database-lock"));
+    m_ui->actionLockDatabaseToolbar->setIcon(icons()->icon("database-lock"));
+    m_ui->actionLockAllDatabases->setIcon(icons()->icon("database-lock-all"));
+    m_ui->actionQuit->setIcon(icons()->icon("application-exit"));
+    m_ui->actionDatabaseMerge->setIcon(icons()->icon("database-merge"));
+    m_ui->menuImport->setIcon(icons()->icon("document-import"));
+    m_ui->menuExport->setIcon(icons()->icon("document-export"));
 
-    m_ui->actionEntryNew->setIcon(resources()->icon("entry-new"));
-    m_ui->actionEntryClone->setIcon(resources()->icon("entry-clone"));
-    m_ui->actionEntryEdit->setIcon(resources()->icon("entry-edit"));
-    m_ui->actionEntryDelete->setIcon(resources()->icon("entry-delete"));
-    m_ui->actionEntryAutoType->setIcon(resources()->icon("auto-type"));
-    m_ui->menuEntryAutoTypeWithSequence->setIcon(resources()->icon("auto-type"));
-    m_ui->actionEntryAutoTypeUsername->setIcon(resources()->icon("auto-type"));
-    m_ui->actionEntryAutoTypeUsernameEnter->setIcon(resources()->icon("auto-type"));
-    m_ui->actionEntryAutoTypePassword->setIcon(resources()->icon("auto-type"));
-    m_ui->actionEntryAutoTypePasswordEnter->setIcon(resources()->icon("auto-type"));
-    m_ui->actionEntryMoveUp->setIcon(resources()->icon("move-up"));
-    m_ui->actionEntryMoveDown->setIcon(resources()->icon("move-down"));
-    m_ui->actionEntryCopyUsername->setIcon(resources()->icon("username-copy"));
-    m_ui->actionEntryCopyPassword->setIcon(resources()->icon("password-copy"));
-    m_ui->actionEntryCopyURL->setIcon(resources()->icon("url-copy"));
-    m_ui->actionEntryDownloadIcon->setIcon(resources()->icon("favicon-download"));
-    m_ui->actionGroupSortAsc->setIcon(resources()->icon("sort-alphabetical-ascending"));
-    m_ui->actionGroupSortDesc->setIcon(resources()->icon("sort-alphabetical-descending"));
+    m_ui->actionEntryNew->setIcon(icons()->icon("entry-new"));
+    m_ui->actionEntryClone->setIcon(icons()->icon("entry-clone"));
+    m_ui->actionEntryEdit->setIcon(icons()->icon("entry-edit"));
+    m_ui->actionEntryDelete->setIcon(icons()->icon("entry-delete"));
+    m_ui->actionEntryRestore->setIcon(icons()->icon("entry-restore"));
+    m_ui->actionEntryAutoType->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryAutoTypeSequence->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryAutoTypeUsername->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryAutoTypeUsernameEnter->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryAutoTypePassword->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryAutoTypePasswordEnter->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryAutoTypeTOTP->setIcon(icons()->icon("auto-type"));
+    m_ui->actionEntryMoveUp->setIcon(icons()->icon("move-up"));
+    m_ui->actionEntryMoveDown->setIcon(icons()->icon("move-down"));
+    m_ui->actionEntryCopyUsername->setIcon(icons()->icon("username-copy"));
+    m_ui->actionEntryCopyPassword->setIcon(icons()->icon("password-copy"));
+    m_ui->actionEntryCopyURL->setIcon(icons()->icon("url-copy"));
+    m_ui->actionEntryDownloadIcon->setIcon(icons()->icon("favicon-download"));
+    m_ui->actionGroupSortAsc->setIcon(icons()->icon("sort-alphabetical-ascending"));
+    m_ui->actionGroupSortDesc->setIcon(icons()->icon("sort-alphabetical-descending"));
 
-    m_ui->actionGroupNew->setIcon(resources()->icon("group-new"));
-    m_ui->actionGroupEdit->setIcon(resources()->icon("group-edit"));
-    m_ui->actionGroupDelete->setIcon(resources()->icon("group-delete"));
-    m_ui->actionGroupEmptyRecycleBin->setIcon(resources()->icon("group-empty-trash"));
-    m_ui->actionEntryOpenUrl->setIcon(resources()->icon("web"));
-    m_ui->actionGroupDownloadFavicons->setIcon(resources()->icon("favicon-download"));
+    m_ui->actionGroupNew->setIcon(icons()->icon("group-new"));
+    m_ui->actionGroupEdit->setIcon(icons()->icon("group-edit"));
+    m_ui->actionGroupClone->setIcon(icons()->icon("group-clone"));
+    m_ui->actionGroupDelete->setIcon(icons()->icon("group-delete"));
+    m_ui->actionGroupEmptyRecycleBin->setIcon(icons()->icon("group-empty-trash"));
+    m_ui->actionEntryOpenUrl->setIcon(icons()->icon("web"));
+    m_ui->actionGroupDownloadFavicons->setIcon(icons()->icon("favicon-download"));
 
-    m_ui->actionSettings->setIcon(resources()->icon("configure"));
-    m_ui->actionPasswordGenerator->setIcon(resources()->icon("password-generator"));
+    m_ui->actionSettings->setIcon(icons()->icon("configure"));
+    m_ui->actionPasswordGenerator->setIcon(icons()->icon("password-generator"));
 
-    m_ui->actionAbout->setIcon(resources()->icon("help-about"));
-    m_ui->actionDonate->setIcon(resources()->icon("donate"));
-    m_ui->actionBugReport->setIcon(resources()->icon("bugreport"));
-    m_ui->actionGettingStarted->setIcon(resources()->icon("getting-started"));
-    m_ui->actionUserGuide->setIcon(resources()->icon("user-guide"));
-    m_ui->actionOnlineHelp->setIcon(resources()->icon("system-help"));
-    m_ui->actionKeyboardShortcuts->setIcon(resources()->icon("keyboard-shortcuts"));
-    m_ui->actionCheckForUpdates->setIcon(resources()->icon("system-software-update"));
+    m_ui->actionAbout->setIcon(icons()->icon("help-about"));
+    m_ui->actionDonate->setIcon(icons()->icon("donate"));
+    m_ui->actionBugReport->setIcon(icons()->icon("bugreport"));
+    m_ui->actionGettingStarted->setIcon(icons()->icon("getting-started"));
+    m_ui->actionUserGuide->setIcon(icons()->icon("user-guide"));
+    m_ui->actionOnlineHelp->setIcon(icons()->icon("system-help"));
+    m_ui->actionKeyboardShortcuts->setIcon(icons()->icon("keyboard-shortcuts"));
+    m_ui->actionCheckForUpdates->setIcon(icons()->icon("system-software-update"));
 
     m_actionMultiplexer.connect(
         SIGNAL(currentModeChanged(DatabaseWidget::Mode)), this, SLOT(setMenuActionState(DatabaseWidget::Mode)));
@@ -407,7 +427,7 @@ MainWindow::MainWindow()
 
     // Notify search when the active database changes or gets locked
     connect(m_ui->tabWidget,
-            SIGNAL(activateDatabaseChanged(DatabaseWidget*)),
+            SIGNAL(activeDatabaseChanged(DatabaseWidget*)),
             m_searchWidget,
             SLOT(databaseChanged(DatabaseWidget*)));
     connect(m_ui->tabWidget, SIGNAL(databaseLocked(DatabaseWidget*)), m_searchWidget, SLOT(databaseChanged()));
@@ -442,13 +462,17 @@ MainWindow::MainWindow()
     connect(m_ui->actionImportOpVault, SIGNAL(triggered()), m_ui->tabWidget, SLOT(importOpVaultDatabase()));
     connect(m_ui->actionExportCsv, SIGNAL(triggered()), m_ui->tabWidget, SLOT(exportToCsv()));
     connect(m_ui->actionExportHtml, SIGNAL(triggered()), m_ui->tabWidget, SLOT(exportToHtml()));
-    connect(m_ui->actionLockDatabases, SIGNAL(triggered()), m_ui->tabWidget, SLOT(lockDatabases()));
+    connect(
+        m_ui->actionLockDatabase, SIGNAL(triggered()), m_ui->tabWidget, SLOT(lockAndSwitchToFirstUnlockedDatabase()));
+    connect(m_ui->actionLockDatabaseToolbar, SIGNAL(triggered()), m_ui->actionLockDatabase, SIGNAL(triggered()));
+    connect(m_ui->actionLockAllDatabases, SIGNAL(triggered()), m_ui->tabWidget, SLOT(lockDatabases()));
     connect(m_ui->actionQuit, SIGNAL(triggered()), SLOT(appExit()));
 
     m_actionMultiplexer.connect(m_ui->actionEntryNew, SIGNAL(triggered()), SLOT(createEntry()));
     m_actionMultiplexer.connect(m_ui->actionEntryClone, SIGNAL(triggered()), SLOT(cloneEntry()));
     m_actionMultiplexer.connect(m_ui->actionEntryEdit, SIGNAL(triggered()), SLOT(switchToEntryEdit()));
     m_actionMultiplexer.connect(m_ui->actionEntryDelete, SIGNAL(triggered()), SLOT(deleteSelectedEntries()));
+    m_actionMultiplexer.connect(m_ui->actionEntryRestore, SIGNAL(triggered()), SLOT(restoreSelectedEntries()));
 
     m_actionMultiplexer.connect(m_ui->actionEntryTotp, SIGNAL(triggered()), SLOT(showTotp()));
     m_actionMultiplexer.connect(m_ui->actionEntrySetupTotp, SIGNAL(triggered()), SLOT(setupTotp()));
@@ -463,6 +487,7 @@ MainWindow::MainWindow()
     m_actionMultiplexer.connect(m_ui->actionEntryCopyURL, SIGNAL(triggered()), SLOT(copyURL()));
     m_actionMultiplexer.connect(m_ui->actionEntryCopyNotes, SIGNAL(triggered()), SLOT(copyNotes()));
     m_actionMultiplexer.connect(m_ui->actionEntryAutoType, SIGNAL(triggered()), SLOT(performAutoType()));
+    m_actionMultiplexer.connect(m_ui->actionEntryAutoTypeSequence, SIGNAL(triggered()), SLOT(performAutoType()));
     m_actionMultiplexer.connect(
         m_ui->actionEntryAutoTypeUsername, SIGNAL(triggered()), SLOT(performAutoTypeUsername()));
     m_actionMultiplexer.connect(
@@ -471,6 +496,7 @@ MainWindow::MainWindow()
         m_ui->actionEntryAutoTypePassword, SIGNAL(triggered()), SLOT(performAutoTypePassword()));
     m_actionMultiplexer.connect(
         m_ui->actionEntryAutoTypePasswordEnter, SIGNAL(triggered()), SLOT(performAutoTypePasswordEnter()));
+    m_actionMultiplexer.connect(m_ui->actionEntryAutoTypeTOTP, SIGNAL(triggered()), SLOT(performAutoTypeTOTP()));
     m_actionMultiplexer.connect(m_ui->actionEntryOpenUrl, SIGNAL(triggered()), SLOT(openUrl()));
     m_actionMultiplexer.connect(m_ui->actionEntryDownloadIcon, SIGNAL(triggered()), SLOT(downloadSelectedFavicons()));
 #ifdef WITH_XC_SSHAGENT
@@ -480,6 +506,7 @@ MainWindow::MainWindow()
 
     m_actionMultiplexer.connect(m_ui->actionGroupNew, SIGNAL(triggered()), SLOT(createGroup()));
     m_actionMultiplexer.connect(m_ui->actionGroupEdit, SIGNAL(triggered()), SLOT(switchToGroupEdit()));
+    m_actionMultiplexer.connect(m_ui->actionGroupClone, SIGNAL(triggered()), SLOT(cloneGroup()));
     m_actionMultiplexer.connect(m_ui->actionGroupDelete, SIGNAL(triggered()), SLOT(deleteGroup()));
     m_actionMultiplexer.connect(m_ui->actionGroupEmptyRecycleBin, SIGNAL(triggered()), SLOT(emptyRecycleBin()));
     m_actionMultiplexer.connect(m_ui->actionGroupSortAsc, SIGNAL(triggered()), SLOT(sortGroupsAsc()));
@@ -589,7 +616,7 @@ MainWindow::MainWindow()
 #if defined(KEEPASSXC_BUILD_TYPE_SNAPSHOT)
     if (config()->get(Config::Messages_HidePreReleaseWarning) != KEEPASSXC_VERSION) {
         m_ui->globalMessageWidget->showMessage(
-            tr("WARNING: You are using an unstable build of KeePassXC!\n"
+            tr("WARNING: You are using an unstable build of KeePassXC.\n"
                "There is a high risk of corruption, maintain a backup of your databases.\n"
                "This version is not meant for production use."),
             MessageWidget::Warning,
@@ -598,15 +625,15 @@ MainWindow::MainWindow()
 #elif defined(KEEPASSXC_BUILD_TYPE_PRE_RELEASE)
     if (config()->get(Config::Messages_HidePreReleaseWarning) != KEEPASSXC_VERSION) {
         m_ui->globalMessageWidget->showMessage(
-            tr("NOTE: You are using a pre-release version of KeePassXC!\n"
-               "Expect some bugs and minor issues, this version is not meant for production use."),
+            tr("NOTE: You are using a pre-release version of KeePassXC.\n"
+               "Expect some bugs and minor issues, this version is meant for testing purposes."),
             MessageWidget::Information,
             -1);
     }
 #elif (QT_VERSION >= QT_VERSION_CHECK(5, 5, 0) && QT_VERSION < QT_VERSION_CHECK(5, 6, 0))
     if (!config()->get(Config::Messages_Qt55CompatibilityWarning).toBool()) {
         m_ui->globalMessageWidget->showMessage(
-            tr("WARNING: Your Qt version may cause KeePassXC to crash with an On-Screen Keyboard!\n"
+            tr("WARNING: Your Qt version may cause KeePassXC to crash with an On-Screen Keyboard.\n"
                "We recommend you use the AppImage available on our downloads page."),
             MessageWidget::Warning,
             -1);
@@ -614,10 +641,10 @@ MainWindow::MainWindow()
     }
 #endif
 
-    QObject::connect(qApp, SIGNAL(anotherInstanceStarted()), this, SLOT(bringToFront()));
-    QObject::connect(qApp, SIGNAL(applicationActivated()), this, SLOT(bringToFront()));
-    QObject::connect(qApp, SIGNAL(openFile(QString)), this, SLOT(openDatabase(QString)));
-    QObject::connect(qApp, SIGNAL(quitSignalReceived()), this, SLOT(appExit()), Qt::DirectConnection);
+    connect(qApp, SIGNAL(anotherInstanceStarted()), this, SLOT(bringToFront()));
+    connect(qApp, SIGNAL(applicationActivated()), this, SLOT(bringToFront()));
+    connect(qApp, SIGNAL(openFile(QString)), this, SLOT(openDatabase(QString)));
+    connect(qApp, SIGNAL(quitSignalReceived()), this, SLOT(appExit()), Qt::DirectConnection);
 
     statusBar()->setFixedHeight(24);
     m_progressBarLabel = new QLabel(statusBar());
@@ -631,6 +658,8 @@ MainWindow::MainWindow()
     m_progressBar->setMaximum(100);
     statusBar()->addPermanentWidget(m_progressBar);
     connect(clipboard(), SIGNAL(updateCountdown(int, QString)), this, SLOT(updateProgressBar(int, QString)));
+
+    restoreConfigState();
 }
 
 MainWindow::~MainWindow()
@@ -640,6 +669,32 @@ MainWindow::~MainWindow()
 #endif
 }
 
+/**
+ * Restore the main window's state after launch
+ */
+void MainWindow::restoreConfigState()
+{
+    // start minimized if configured
+    if (config()->get(Config::GUI_MinimizeOnStartup).toBool()) {
+        hideWindow();
+    } else {
+        bringToFront();
+    }
+
+    if (config()->get(Config::OpenPreviousDatabasesOnStartup).toBool()) {
+        const QStringList fileNames = config()->get(Config::LastOpenedDatabases).toStringList();
+        for (const QString& filename : fileNames) {
+            if (!filename.isEmpty() && QFile::exists(filename)) {
+                openDatabase(filename);
+            }
+        }
+        auto lastActiveFile = config()->get(Config::LastActiveDatabase).toString();
+        if (!lastActiveFile.isEmpty()) {
+            openDatabase(lastActiveFile);
+        }
+    }
+}
+
 QList<DatabaseWidget*> MainWindow::getOpenDatabases()
 {
     QList<DatabaseWidget*> dbWidgets;
@@ -731,7 +786,9 @@ void MainWindow::setMenuActionState(DatabaseWidget::Mode mode)
     m_ui->actionDatabaseOpen->setEnabled(inDatabaseTabWidgetOrWelcomeWidget);
     m_ui->menuRecentDatabases->setEnabled(inDatabaseTabWidgetOrWelcomeWidget);
     m_ui->menuImport->setEnabled(inDatabaseTabWidgetOrWelcomeWidget);
-    m_ui->actionLockDatabases->setEnabled(m_ui->tabWidget->hasLockableDatabases());
+    m_ui->actionLockDatabase->setEnabled(m_ui->tabWidget->hasLockableDatabases());
+    m_ui->actionLockDatabaseToolbar->setEnabled(m_ui->tabWidget->hasLockableDatabases());
+    m_ui->actionLockAllDatabases->setEnabled(m_ui->tabWidget->hasLockableDatabases());
 
     if (inDatabaseTabWidget && m_ui->tabWidget->currentIndex() != -1) {
         DatabaseWidget* dbWidget = m_ui->tabWidget->currentDatabaseWidget();
@@ -757,6 +814,10 @@ void MainWindow::setMenuActionState(DatabaseWidget::Mode mode)
             m_ui->actionEntryClone->setEnabled(singleEntrySelected);
             m_ui->actionEntryEdit->setEnabled(singleEntrySelected);
             m_ui->actionEntryDelete->setEnabled(entriesSelected);
+            m_ui->actionEntryRestore->setVisible(entriesSelected && recycleBinSelected);
+            m_ui->actionEntryRestore->setEnabled(entriesSelected && recycleBinSelected);
+            m_ui->actionEntryRestore->setText(tr("Restore Entry(s)", "", dbWidget->numberOfSelectedEntries()));
+            m_ui->actionEntryRestore->setToolTip(tr("Restore Entry(s)", "", dbWidget->numberOfSelectedEntries()));
             m_ui->actionEntryMoveUp->setVisible(!sorted);
             m_ui->actionEntryMoveDown->setVisible(!sorted);
             m_ui->actionEntryMoveUp->setEnabled(singleEntrySelected && !sorted && entryIndex > 0);
@@ -772,13 +833,19 @@ void MainWindow::setMenuActionState(DatabaseWidget::Mode mode)
             m_ui->menuEntryCopyAttribute->setEnabled(singleEntrySelected);
             m_ui->menuEntryTotp->setEnabled(singleEntrySelected);
             m_ui->actionEntryAutoType->setEnabled(singleEntrySelected);
-            m_ui->menuEntryAutoTypeWithSequence->setEnabled(singleEntrySelected);
+            m_ui->actionEntryAutoType->menu()->setEnabled(singleEntrySelected);
+            m_ui->actionEntryAutoTypeSequence->setText(
+                singleEntrySelected ? dbWidget->currentSelectedEntry()->effectiveAutoTypeSequence()
+                                    : Group::RootAutoTypeSequence);
+            m_ui->actionEntryAutoTypeSequence->setEnabled(singleEntrySelected);
             m_ui->actionEntryAutoTypeUsername->setEnabled(singleEntrySelected && dbWidget->currentEntryHasUsername());
             m_ui->actionEntryAutoTypeUsernameEnter->setEnabled(singleEntrySelected
                                                                && dbWidget->currentEntryHasUsername());
             m_ui->actionEntryAutoTypePassword->setEnabled(singleEntrySelected && dbWidget->currentEntryHasPassword());
             m_ui->actionEntryAutoTypePasswordEnter->setEnabled(singleEntrySelected
                                                                && dbWidget->currentEntryHasPassword());
+            m_ui->actionEntryAutoTypeTOTP->setEnabled(singleEntrySelected && dbWidget->currentEntryHasTotp());
+            m_ui->actionEntryAutoTypeTOTP->setVisible(singleEntrySelected && dbWidget->currentEntryHasTotp());
             m_ui->actionEntryOpenUrl->setEnabled(singleEntrySelected && dbWidget->currentEntryHasUrl());
             m_ui->actionEntryTotp->setEnabled(singleEntrySelected && dbWidget->currentEntryHasTotp());
             m_ui->actionEntryCopyTotp->setEnabled(singleEntrySelected && dbWidget->currentEntryHasTotp());
@@ -788,6 +855,7 @@ void MainWindow::setMenuActionState(DatabaseWidget::Mode mode)
                                                       || (singleEntrySelected && dbWidget->currentEntryHasUrl()));
             m_ui->actionGroupNew->setEnabled(groupSelected);
             m_ui->actionGroupEdit->setEnabled(groupSelected);
+            m_ui->actionGroupClone->setEnabled(groupSelected && dbWidget->canCloneCurrentGroup());
             m_ui->actionGroupDelete->setEnabled(groupSelected && dbWidget->canDeleteCurrentGroup());
             m_ui->actionGroupSortAsc->setEnabled(groupSelected && currentGroupHasChildren);
             m_ui->actionGroupSortDesc->setEnabled(groupSelected && currentGroupHasChildren);
@@ -831,7 +899,6 @@ void MainWindow::setMenuActionState(DatabaseWidget::Mode mode)
                                                                m_ui->actionEntryCopyURL,
                                                                m_ui->actionEntryOpenUrl,
                                                                m_ui->actionEntryAutoType,
-                                                               m_ui->menuEntryAutoTypeWithSequence->menuAction(),
                                                                m_ui->actionEntryDownloadIcon,
                                                                m_ui->actionEntryCopyNotes,
                                                                m_ui->actionEntryCopyTitle,
@@ -865,6 +932,9 @@ void MainWindow::setMenuActionState(DatabaseWidget::Mode mode)
             m_ui->actionExportCsv->setEnabled(false);
             m_ui->actionExportHtml->setEnabled(false);
             m_ui->actionDatabaseMerge->setEnabled(false);
+            // Only disable the action in the database menu so that the
+            // menu remains active in the toolbar, if necessary
+            m_ui->actionLockDatabase->setEnabled(false);
 
             m_searchWidgetAction->setEnabled(false);
             break;
@@ -1258,7 +1328,11 @@ bool MainWindow::focusNextPrevChild(bool next)
         // Search Widget <-> Tab Widget <-> DbWidget
         if (next) {
             if (m_searchWidget->hasFocus()) {
-                m_ui->tabWidget->setFocus(Qt::TabFocusReason);
+                if (m_ui->tabWidget->count() > 1) {
+                    m_ui->tabWidget->setFocus(Qt::TabFocusReason);
+                } else {
+                    dbWidget->setFocus(Qt::TabFocusReason);
+                }
             } else if (m_ui->tabWidget->hasFocus()) {
                 dbWidget->setFocus(Qt::TabFocusReason);
             } else {
@@ -1270,7 +1344,11 @@ bool MainWindow::focusNextPrevChild(bool next)
             } else if (m_ui->tabWidget->hasFocus()) {
                 focusSearchWidget();
             } else {
-                m_ui->tabWidget->setFocus(Qt::BacktabFocusReason);
+                if (m_ui->tabWidget->count() > 1) {
+                    m_ui->tabWidget->setFocus(Qt::BacktabFocusReason);
+                } else {
+                    focusSearchWidget();
+                }
             }
         }
         return true;
@@ -1331,9 +1409,9 @@ void MainWindow::updateTrayIcon()
 
             auto* actionToggle = new QAction(tr("Toggle window"), menu);
             menu->addAction(actionToggle);
-            actionToggle->setIcon(resources()->icon("keepassxc-monochrome-dark"));
+            actionToggle->setIcon(icons()->icon("keepassxc-monochrome-dark"));
 
-            menu->addAction(m_ui->actionLockDatabases);
+            menu->addAction(m_ui->actionLockAllDatabases);
 
 #ifdef Q_OS_MACOS
             auto actionQuit = new QAction(tr("Quit KeePassXC"), menu);
@@ -1351,9 +1429,9 @@ void MainWindow::updateTrayIcon()
         }
 
         if (m_ui->tabWidget->hasLockableDatabases()) {
-            m_trayIcon->setIcon(resources()->trayIconUnlocked());
+            m_trayIcon->setIcon(icons()->trayIconUnlocked());
         } else {
-            m_trayIcon->setIcon(resources()->trayIconLocked());
+            m_trayIcon->setIcon(icons()->trayIconLocked());
         }
 
         m_trayIcon->setToolTip(windowTitle().replace("[*]", isWindowModified() ? "*" : ""));
@@ -1450,21 +1528,6 @@ void MainWindow::applySettingsChanges()
         m_inactivityTimer->deactivate();
     }
 
-#ifdef WITH_XC_TOUCHID
-    if (config()->get(Config::Security_ResetTouchId).toBool()) {
-        // Calculate TouchID timeout in milliseconds
-        timeout = config()->get(Config::Security_ResetTouchIdTimeout).toInt() * 60 * 1000;
-        if (timeout <= 0) {
-            timeout = 30 * 60 * 1000;
-        }
-
-        m_touchIDinactivityTimer->setInactivityTimeout(timeout);
-        m_touchIDinactivityTimer->activate();
-    } else {
-        m_touchIDinactivityTimer->deactivate();
-    }
-#endif
-
     m_ui->toolBar->setHidden(config()->get(Config::GUI_HideToolbar).toBool());
     m_ui->toolBar->setMovable(config()->get(Config::GUI_MovableToolbar).toBool());
 
@@ -1616,18 +1679,18 @@ void MainWindow::toggleWindow()
     }
 }
 
+void MainWindow::closeModalWindow()
+{
+    if (qApp->modalWindow()) {
+        qApp->modalWindow()->close();
+    }
+}
+
 void MainWindow::lockDatabasesAfterInactivity()
 {
     m_ui->tabWidget->lockDatabases();
 }
 
-void MainWindow::forgetTouchIDAfterInactivity()
-{
-#ifdef WITH_XC_TOUCHID
-    TouchID::getInstance().reset();
-#endif
-}
-
 bool MainWindow::isTrayIconEnabled() const
 {
     return m_trayIcon && m_trayIcon->isVisible();
@@ -1657,7 +1720,7 @@ void MainWindow::hideGlobalMessage()
 
 void MainWindow::showYubiKeyPopup()
 {
-    displayGlobalMessage(tr("Please touch the button on your YubiKey!"),
+    displayGlobalMessage(tr("Please present or touch your YubiKey to continue…"),
                          MessageWidget::Information,
                          false,
                          MessageWidget::DisableAutoHide);
@@ -1684,12 +1747,6 @@ void MainWindow::handleScreenLock()
     if (config()->get(Config::Security_LockDatabaseScreenLock).toBool()) {
         lockDatabasesAfterInactivity();
     }
-
-#ifdef WITH_XC_TOUCHID
-    if (config()->get(Config::Security_ResetTouchIdScreenlock).toBool()) {
-        forgetTouchIDAfterInactivity();
-    }
-#endif
 }
 
 QStringList MainWindow::kdbxFilesFromUrls(const QList<QUrl>& urls)
@@ -1752,7 +1809,7 @@ void MainWindow::displayDesktopNotification(const QString& msg, QString title, i
     }
 
 #if QT_VERSION >= QT_VERSION_CHECK(5, 9, 0)
-    m_trayIcon->showMessage(title, msg, resources()->applicationIcon(), msTimeoutHint);
+    m_trayIcon->showMessage(title, msg, icons()->applicationIcon(), msTimeoutHint);
 #else
     m_trayIcon->showMessage(title, msg, QSystemTrayIcon::Information, msTimeoutHint);
 #endif
