@@ -1,6 +1,6 @@
 /*
  *  Copyright (C) 2010 Felix Geyer <debfx@fobos.de>
- *  Copyright (C) 2017 KeePassXC Team <team@keepassxc.org>
+ *  Copyright (C) 2020 KeePassXC Team <team@keepassxc.org>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -18,17 +18,16 @@
 
 #include <QCommandLineParser>
 #include <QFile>
-#include <QTextStream>
+#include <QWindow>
 
 #include "cli/Utils.h"
 #include "config-keepassx.h"
-#include "core/Bootstrap.h"
-#include "core/Config.h"
 #include "core/Tools.h"
 #include "crypto/Crypto.h"
 #include "gui/Application.h"
 #include "gui/MainWindow.h"
 #include "gui/MessageBox.h"
+#include "gui/osutils/OSUtils.h"
 
 #if defined(WITH_ASAN) && defined(WITH_LSAN)
 #include <sanitizer/lsan_interface.h>
@@ -55,6 +54,12 @@ int main(int argc, char** argv)
 #if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0) && defined(Q_OS_WIN)
     QGuiApplication::setHighDpiScaleFactorRoundingPolicy(Qt::HighDpiScaleFactorRoundingPolicy::PassThrough);
 #endif
+    Application app(argc, argv);
+    // don't set organizationName as that changes the return value of
+    // QStandardPaths::writableLocation(QDesktopServices::DataLocation)
+    Application::setApplicationName("KeePassXC");
+    Application::setApplicationVersion(KEEPASSXC_VERSION);
+    app.setProperty("KPXC_QUALIFIED_APPNAME", "org.keepassxc.KeePassXC");
 
     QCommandLineParser parser;
     parser.setApplicationDescription(QObject::tr("KeePassXC - cross-platform password manager"));
@@ -67,6 +72,8 @@ int main(int argc, char** argv)
     QCommandLineOption lockOption("lock", QObject::tr("lock all open databases"));
     QCommandLineOption keyfileOption("keyfile", QObject::tr("key file of the database"), "keyfile");
     QCommandLineOption pwstdinOption("pw-stdin", QObject::tr("read password of the database from stdin"));
+    QCommandLineOption allowScreenCaptureOption("allow-screencapture",
+                                                QObject::tr("allow app screen recordering and screenshots"));
 
     QCommandLineOption helpOption = parser.addHelpOption();
     QCommandLineOption versionOption = parser.addVersionOption();
@@ -78,12 +85,9 @@ int main(int argc, char** argv)
     parser.addOption(pwstdinOption);
     parser.addOption(debugInfoOption);
 
-    Application app(argc, argv);
-    // don't set organizationName as that changes the return value of
-    // QStandardPaths::writableLocation(QDesktopServices::DataLocation)
-    Application::setApplicationName("KeePassXC");
-    Application::setApplicationVersion(KEEPASSXC_VERSION);
-    app.setProperty("KPXC_QUALIFIED_APPNAME", "org.keepassxc.KeePassXC");
+    if (osUtils->canPreventScreenCapture()) {
+        parser.addOption(allowScreenCaptureOption);
+    }
 
     parser.process(app);
 
@@ -92,6 +96,14 @@ int main(int argc, char** argv)
         return EXIT_SUCCESS;
     }
 
+    // Show debug information and then exit
+    if (parser.isSet(debugInfoOption)) {
+        QTextStream out(stdout, QIODevice::WriteOnly);
+        QString debugInfo = Tools::debugInfo().append("\n").append(Crypto::debugInfo());
+        out << debugInfo << endl;
+        return EXIT_SUCCESS;
+    }
+
     // Process config file options early
     if (parser.isSet(configOption) || parser.isSet(localConfigOption)) {
         Config::createConfigFromFile(parser.value(configOption), parser.value(localConfigOption));
@@ -118,15 +130,6 @@ int main(int argc, char** argv)
         return EXIT_SUCCESS;
     }
 
-    // Apply the configured theme before creating any GUI elements
-    app.applyTheme();
-
-#if QT_VERSION >= QT_VERSION_CHECK(5, 7, 0)
-    QGuiApplication::setDesktopFileName(app.property("KPXC_QUALIFIED_APPNAME").toString() + QStringLiteral(".desktop"));
-#endif
-
-    Bootstrap::bootstrapApplication();
-
     if (!Crypto::init()) {
         QString error = QObject::tr("Fatal error while testing the cryptographic functions.");
         error.append("\n");
@@ -135,24 +138,37 @@ int main(int argc, char** argv)
         return EXIT_FAILURE;
     }
 
-    // Displaying the debugging informations must be done after Crypto::init,
-    // to make sure we know which libgcrypt version is used.
-    if (parser.isSet(debugInfoOption)) {
-        QTextStream out(stdout, QIODevice::WriteOnly);
-        QString debugInfo = Tools::debugInfo().append("\n").append(Crypto::debugInfo());
-        out << debugInfo << endl;
-        return EXIT_SUCCESS;
-    }
+    // Apply the configured theme before creating any GUI elements
+    app.applyTheme();
+
+#if QT_VERSION >= QT_VERSION_CHECK(5, 7, 0)
+    QGuiApplication::setDesktopFileName(app.property("KPXC_QUALIFIED_APPNAME").toString() + QStringLiteral(".desktop"));
+#endif
+
+    Application::bootstrap();
 
     MainWindow mainWindow;
-    QObject::connect(&app, SIGNAL(anotherInstanceStarted()), &mainWindow, SLOT(bringToFront()));
-    QObject::connect(&app, SIGNAL(applicationActivated()), &mainWindow, SLOT(bringToFront()));
-    QObject::connect(&app, SIGNAL(openFile(QString)), &mainWindow, SLOT(openDatabase(QString)));
-    QObject::connect(&app, SIGNAL(quitSignalReceived()), &mainWindow, SLOT(appExit()), Qt::DirectConnection);
 
-    Bootstrap::restoreMainWindowState(mainWindow);
+#ifndef QT_DEBUG
+    // Disable screen capture if capable and not explicitly allowed
+    if (osUtils->canPreventScreenCapture() && !parser.isSet(allowScreenCaptureOption)) {
+        // This ensures any top-level windows (Main Window, Modal Dialogs, etc.) are excluded from screenshots
+        QObject::connect(&app, &QGuiApplication::focusWindowChanged, &mainWindow, [&](QWindow* window) {
+            if (window) {
+                if (!osUtils->setPreventScreenCapture(window, true)) {
+                    mainWindow.displayGlobalMessage(
+                        QObject::tr("Warning: Failed to prevent screenshots on a top level window!"),
+                        MessageWidget::Error);
+                }
+            }
+        });
+    }
+#endif
 
     const bool pwstdin = parser.isSet(pwstdinOption);
+    if (!fileNames.isEmpty() && pwstdin) {
+        Utils::setDefaultTextStreams();
+    }
     for (const QString& filename : fileNames) {
         QString password;
         if (pwstdin) {
@@ -160,7 +176,6 @@ int main(int argc, char** argv)
             // buffer for native messaging, even if the specified file does not exist
             QTextStream out(stdout, QIODevice::WriteOnly);
             out << QObject::tr("Database password: ") << flush;
-            Utils::setDefaultTextStreams();
             password = Utils::getPassword();
         }
 
