@@ -17,33 +17,33 @@
  */
 
 #include "OpenSSHKey.h"
+
 #include "ASN1Key.h"
-#include <QRegularExpression>
-#include <QStringList>
-#include <QCryptographicHash>
+#include "BinaryStream.h"
 #include "crypto/SymmetricCipher.h"
 
-const QString OpenSSHKey::TYPE_DSA = "DSA PRIVATE KEY";
-const QString OpenSSHKey::TYPE_RSA = "RSA PRIVATE KEY";
-const QString OpenSSHKey::TYPE_OPENSSH = "OPENSSH PRIVATE KEY";
+#include <QRegularExpression>
+#include <QStringList>
 
-// bcrypt_pbkdf.cpp
-int bcrypt_pbkdf(const QByteArray& pass, const QByteArray& salt, QByteArray& key, quint32 rounds);
+#include <botan/pwdhash.h>
 
-OpenSSHKey::OpenSSHKey(QObject *parent)
+const QString OpenSSHKey::TYPE_DSA_PRIVATE = "DSA PRIVATE KEY";
+const QString OpenSSHKey::TYPE_RSA_PRIVATE = "RSA PRIVATE KEY";
+const QString OpenSSHKey::TYPE_OPENSSH_PRIVATE = "OPENSSH PRIVATE KEY";
+
+OpenSSHKey::OpenSSHKey(QObject* parent)
     : QObject(parent)
     , m_type(QString())
     , m_cipherName(QString("none"))
     , m_kdfName(QString("none"))
     , m_kdfOptions(QByteArray())
+    , m_rawType(QString())
+    , m_rawData(QByteArray())
+    , m_rawPublicData(QByteArray())
     , m_rawPrivateData(QByteArray())
-    , m_publicData(QList<QByteArray>())
-    , m_privateData(QList<QByteArray>())
-    , m_privateType(QString())
     , m_comment(QString())
     , m_error(QString())
 {
-
 }
 
 OpenSSHKey::OpenSSHKey(const OpenSSHKey& other)
@@ -52,13 +52,13 @@ OpenSSHKey::OpenSSHKey(const OpenSSHKey& other)
     , m_cipherName(other.m_cipherName)
     , m_kdfName(other.m_kdfName)
     , m_kdfOptions(other.m_kdfOptions)
+    , m_rawType(other.m_rawType)
+    , m_rawData(other.m_rawData)
+    , m_rawPublicData(other.m_rawPublicData)
     , m_rawPrivateData(other.m_rawPrivateData)
-    , m_publicData(other.m_publicData)
-    , m_privateData(other.m_privateData)
     , m_comment(other.m_comment)
     , m_error(other.m_error)
 {
-
 }
 
 bool OpenSSHKey::operator==(const OpenSSHKey& other) const
@@ -77,24 +77,9 @@ const QString OpenSSHKey::type() const
     return m_type;
 }
 
-int OpenSSHKey::keyLength() const
+const QString OpenSSHKey::fingerprint(QCryptographicHash::Algorithm algo) const
 {
-    if (m_type == "ssh-dss" && m_publicData.length() == 4) {
-        return (m_publicData[0].length() - 1) * 8;
-    } else if (m_type == "ssh-rsa" && m_publicData.length() == 2) {
-        return (m_publicData[1].length() - 1) * 8;
-    } else if (m_type.startsWith("ecdsa-sha2-") && m_publicData.length() == 2) {
-        return (m_publicData[1].length() - 1) * 4;
-    } else if (m_type == "ssh-ed25519" && m_publicData.length() == 1) {
-        return m_publicData[0].length() * 8;
-    }
-
-    return 0;
-}
-
-const QString OpenSSHKey::fingerprint() const
-{
-    if (m_publicData.isEmpty()) {
+    if (m_rawPublicData.isEmpty()) {
         return {};
     }
 
@@ -102,14 +87,22 @@ const QString OpenSSHKey::fingerprint() const
     BinaryStream stream(&publicKey);
 
     stream.writeString(m_type);
+    stream.write(m_rawPublicData);
 
-    for (QByteArray ba : m_publicData) {
-        stream.writeString(ba);
+    QByteArray rawHash = QCryptographicHash::hash(publicKey, algo);
+
+    if (algo == QCryptographicHash::Md5) {
+        QString md5Hash = QString::fromLatin1(rawHash.toHex());
+        QStringList md5HashParts;
+        for (int i = 0; i < md5Hash.length(); i += 2) {
+            md5HashParts.append(md5Hash.mid(i, 2));
+        }
+        return "MD5:" + md5HashParts.join(':');
+    } else if (algo == QCryptographicHash::Sha256) {
+        return "SHA256:" + QString::fromLatin1(rawHash.toBase64(QByteArray::OmitTrailingEquals));
     }
 
-    QByteArray rawHash = QCryptographicHash::hash(publicKey, QCryptographicHash::Sha256);
-
-    return "SHA256:" + QString::fromLatin1(rawHash.toBase64(QByteArray::OmitTrailingEquals));
+    return "HASH:" + QString::fromLatin1(rawHash.toHex());
 }
 
 const QString OpenSSHKey::comment() const
@@ -119,7 +112,7 @@ const QString OpenSSHKey::comment() const
 
 const QString OpenSSHKey::publicKey() const
 {
-    if (m_publicData.isEmpty()) {
+    if (m_rawPublicData.isEmpty()) {
         return {};
     }
 
@@ -127,10 +120,7 @@ const QString OpenSSHKey::publicKey() const
     BinaryStream stream(&publicKey);
 
     stream.writeString(m_type);
-
-    for (QByteArray ba : m_publicData) {
-        stream.writeString(ba);
-    }
+    stream.write(m_rawPublicData);
 
     return m_type + " " + QString::fromLatin1(publicKey.toBase64()) + " " + m_comment;
 }
@@ -145,14 +135,14 @@ void OpenSSHKey::setType(const QString& type)
     m_type = type;
 }
 
-void OpenSSHKey::setPublicData(const QList<QByteArray>& data)
+void OpenSSHKey::setPublicData(const QByteArray& data)
 {
-    m_publicData = data;
+    m_rawPublicData = data;
 }
 
-void OpenSSHKey::setPrivateData(const QList<QByteArray>& data)
+void OpenSSHKey::setPrivateData(const QByteArray& data)
 {
-    m_privateData = data;
+    m_rawPrivateData = data;
 }
 
 void OpenSSHKey::setComment(const QString& comment)
@@ -162,11 +152,11 @@ void OpenSSHKey::setComment(const QString& comment)
 
 void OpenSSHKey::clearPrivate()
 {
+    m_rawData.clear();
     m_rawPrivateData.clear();
-    m_privateData.clear();
 }
 
-bool OpenSSHKey::parsePEM(const QByteArray& in, QByteArray& out)
+bool OpenSSHKey::extractPEM(const QByteArray& in, QByteArray& out)
 {
     QString pem = QString::fromLatin1(in);
     QStringList rows = pem.split(QRegularExpression("(?:\r?\n|\r)"), QString::SkipEmptyParts);
@@ -192,7 +182,7 @@ bool OpenSSHKey::parsePEM(const QByteArray& in, QByteArray& out)
         return false;
     }
 
-    m_privateType = beginMatch.captured(1);
+    m_rawType = beginMatch.captured(1);
 
     rows.removeFirst();
     rows.removeLast();
@@ -228,17 +218,17 @@ bool OpenSSHKey::parsePEM(const QByteArray& in, QByteArray& out)
     return true;
 }
 
-bool OpenSSHKey::parse(const QByteArray& in)
+bool OpenSSHKey::parsePKCS1PEM(const QByteArray& in)
 {
     QByteArray data;
 
-    if (!parsePEM(in, data)) {
+    if (!extractPEM(in, data)) {
         return false;
     }
 
-    if (m_privateType == TYPE_DSA || m_privateType == TYPE_RSA) {
-        m_rawPrivateData = data;
-    } else if (m_privateType == TYPE_OPENSSH) {
+    if (m_rawType == TYPE_DSA_PRIVATE || m_rawType == TYPE_RSA_PRIVATE) {
+        m_rawData = data;
+    } else if (m_rawType == TYPE_OPENSSH_PRIVATE) {
         BinaryStream stream(&data);
 
         QByteArray magic;
@@ -282,18 +272,18 @@ bool OpenSSHKey::parse(const QByteArray& in)
         }
 
         // padded list of keys
-        if (!stream.readString(m_rawPrivateData)) {
+        if (!stream.readString(m_rawData)) {
             m_error = tr("Corrupted key file, reading private key failed");
             return false;
         }
     } else {
-        m_error = tr("Unsupported key type: %1").arg(m_privateType);
+        m_error = tr("Unsupported key type: %1").arg(m_rawType);
         return false;
     }
 
     // load private if no encryption
     if (!encrypted()) {
-        return openPrivateKey();
+        return openKey();
     }
 
     return true;
@@ -304,127 +294,121 @@ bool OpenSSHKey::encrypted() const
     return (m_cipherName != "none");
 }
 
-bool OpenSSHKey::openPrivateKey(const QString& passphrase)
+bool OpenSSHKey::openKey(const QString& passphrase)
 {
-    QScopedPointer<SymmetricCipher> cipher;
+    QScopedPointer<SymmetricCipher> cipher(new SymmetricCipher());
 
-    if (!m_privateData.isEmpty()) {
+    if (!m_rawPrivateData.isEmpty()) {
         return true;
     }
 
-    if (m_rawPrivateData.isEmpty()) {
+    if (m_rawData.isEmpty()) {
         m_error = tr("No private key payload to decrypt");
         return false;
     }
 
-    if (m_cipherName.compare("aes-128-cbc", Qt::CaseInsensitive) == 0) {
-        cipher.reset(new SymmetricCipher(SymmetricCipher::Aes128, SymmetricCipher::Cbc, SymmetricCipher::Decrypt));
-    } else if (m_cipherName == "aes256-cbc" || m_cipherName.compare("aes-256-cbc", Qt::CaseInsensitive) == 0) {
-        cipher.reset(new SymmetricCipher(SymmetricCipher::Aes256, SymmetricCipher::Cbc, SymmetricCipher::Decrypt));
-    } else if (m_cipherName == "aes256-ctr" || m_cipherName.compare("aes-256-ctr", Qt::CaseInsensitive) == 0) {
-        cipher.reset(new SymmetricCipher(SymmetricCipher::Aes256, SymmetricCipher::Ctr, SymmetricCipher::Decrypt));
-    } else if (m_cipherName != "none") {
-        m_error = tr("Unknown cipher: %1").arg(m_cipherName);
-        return false;
-    }
+    QByteArray rawData = m_rawData;
 
-    if (m_kdfName == "bcrypt") {
-        if (!cipher) {
-            m_error = tr("Trying to run KDF without cipher");
-            return false;
-        }
-
-        if (passphrase.isEmpty()) {
-            m_error = tr("Passphrase is required to decrypt this key");
-            return false;
-        }
-
-        BinaryStream optionStream(&m_kdfOptions);
-
-        QByteArray salt;
-        quint32 rounds;
-
-        optionStream.readString(salt);
-        optionStream.read(rounds);
-
-        QByteArray decryptKey;
-        decryptKey.fill(0, cipher->keySize() + cipher->blockSize());
-
-        QByteArray phraseData = passphrase.toUtf8();
-        if (bcrypt_pbkdf(phraseData, salt, decryptKey, rounds) < 0) {
-            m_error = tr("Key derivation failed, key file corrupted?");
+    if (m_cipherName != "none") {
+        auto cipherMode = SymmetricCipher::stringToMode(m_cipherName);
+        if (cipherMode == SymmetricCipher::InvalidMode) {
+            m_error = tr("Unknown cipher: %1").arg(m_cipherName);
             return false;
         }
 
         QByteArray keyData, ivData;
-        keyData.setRawData(decryptKey.data(), cipher->keySize());
-        ivData.setRawData(decryptKey.data() + cipher->keySize(), cipher->blockSize());
 
-        cipher->init(keyData, ivData);
+        if (m_kdfName == "bcrypt") {
+            if (passphrase.isEmpty()) {
+                m_error = tr("Passphrase is required to decrypt this key");
+                return false;
+            }
 
-        if (!cipher->init(keyData, ivData)) {
-            m_error = cipher->errorString();
-            return false;
-        }
-    } else if (m_kdfName == "md5") {
-        if (m_cipherIV.length() < 8) {
-            m_error = tr("Cipher IV is too short for MD5 kdf");
+            int keySize = cipher->keySize(cipherMode);
+            int blockSize = 16;
+
+            BinaryStream optionStream(&m_kdfOptions);
+
+            QByteArray salt;
+            quint32 rounds;
+
+            optionStream.readString(salt);
+            optionStream.read(rounds);
+
+            QByteArray decryptKey(keySize + blockSize, '\0');
+            try {
+                auto baPass = passphrase.toUtf8();
+                auto pwhash = Botan::PasswordHashFamily::create_or_throw("Bcrypt-PBKDF")->from_iterations(rounds);
+                pwhash->derive_key(reinterpret_cast<uint8_t*>(decryptKey.data()),
+                                   decryptKey.size(),
+                                   baPass.constData(),
+                                   baPass.size(),
+                                   reinterpret_cast<const uint8_t*>(salt.constData()),
+                                   salt.size());
+            } catch (std::exception& e) {
+                m_error = tr("Key derivation failed: %1").arg(e.what());
+                return false;
+            }
+
+            keyData = decryptKey.left(keySize);
+            ivData = decryptKey.right(blockSize);
+        } else if (m_kdfName == "md5") {
+            if (m_cipherIV.length() < 8) {
+                m_error = tr("Cipher IV is too short for MD5 kdf");
+                return false;
+            }
+
+            int keySize = cipher->keySize(cipherMode);
+
+            QByteArray mdBuf;
+            do {
+                QCryptographicHash hash(QCryptographicHash::Md5);
+                hash.addData(mdBuf);
+                hash.addData(passphrase.toUtf8());
+                hash.addData(m_cipherIV.data(), 8);
+                mdBuf = hash.result();
+                keyData.append(mdBuf);
+            } while (keyData.size() < keySize);
+
+            if (keyData.size() > keySize) {
+                // If our key size isn't a multiple of 16 (e.g. AES-192 or something),
+                // then we will need to truncate it.
+                keyData.resize(keySize);
+            }
+
+            ivData = m_cipherIV;
+        } else if (m_kdfName != "none") {
+            m_error = tr("Unknown KDF: %1").arg(m_kdfName);
             return false;
         }
 
-        QByteArray keyData;
-        QByteArray mdBuf;
-        do {
-            QCryptographicHash hash(QCryptographicHash::Md5);
-            hash.addData(mdBuf);
-            hash.addData(passphrase.toUtf8());
-            hash.addData(m_cipherIV.data(), 8);
-            mdBuf = hash.result();
-            keyData.append(mdBuf);
-        } while(keyData.size() < cipher->keySize());
-
-        if (keyData.size() > cipher->keySize()) {
-            // If our key size isn't a multiple of 16 (e.g. AES-192 or something),
-            // then we will need to truncate it.
-            keyData.resize(cipher->keySize());
-        }
-
-        if (!cipher->init(keyData, m_cipherIV)) {
-            m_error = cipher->errorString();
+        // Initialize the cipher using the processed key and iv data
+        if (!cipher->init(cipherMode, SymmetricCipher::Decrypt, keyData, ivData)) {
+            m_error = tr("Failed to initialize cipher: %1").arg(cipher->errorString());
             return false;
         }
-    } else if (m_kdfName != "none") {
-        m_error = tr("Unknown KDF: %1").arg(m_kdfName);
-        return false;
-    }
-
-    QByteArray rawPrivateData = m_rawPrivateData;
-
-    if (cipher && cipher->isInitalized()) {
-        bool ok = false;
-        rawPrivateData = cipher->process(rawPrivateData, &ok);
-        if (!ok) {
-            m_error = tr("Decryption failed, wrong passphrase?");
+        // Decrypt the raw data, we do not use finish because padding is handled separately
+        if (!cipher->process(rawData)) {
+            m_error = tr("Decryption failed: %1").arg(cipher->errorString());
             return false;
         }
     }
 
-    if (m_privateType == TYPE_DSA) {
-        if (!ASN1Key::parseDSA(rawPrivateData, *this)) {
+    if (m_rawType == TYPE_DSA_PRIVATE) {
+        if (!ASN1Key::parseDSA(rawData, *this)) {
             m_error = tr("Decryption failed, wrong passphrase?");
             return false;
         }
 
         return true;
-    } else if (m_privateType == TYPE_RSA) {
-        if (!ASN1Key::parseRSA(rawPrivateData, *this)) {
+    } else if (m_rawType == TYPE_RSA_PRIVATE) {
+        if (!ASN1Key::parseRSA(rawData, *this)) {
             m_error = tr("Decryption failed, wrong passphrase?");
             return false;
         }
-
         return true;
-    } else if (m_privateType == TYPE_OPENSSH) {
-        BinaryStream keyStream(&rawPrivateData);
+    } else if (m_rawType == TYPE_OPENSSH_PRIVATE) {
+        BinaryStream keyStream(&rawData);
 
         quint32 checkInt1;
         quint32 checkInt2;
@@ -440,79 +424,107 @@ bool OpenSSHKey::openPrivateKey(const QString& passphrase)
         return readPrivate(keyStream);
     }
 
-    m_error = tr("Unsupported key type: %1").arg(m_privateType);
+    m_error = tr("Unsupported key type: %1").arg(m_rawType);
     return false;
 }
 
+bool OpenSSHKey::readKeyParts(BinaryStream& in, const QList<KeyPart> parts, BinaryStream& out)
+{
+    for (auto part : parts) {
+        switch (part) {
+        case STR_PART: {
+            QByteArray t;
+
+            if (!in.readString(t)) {
+                m_error = tr("Unexpected EOF while reading key");
+                return false;
+            }
+
+            out.writeString(t);
+            break;
+        }
+        case UINT8_PART: {
+            quint8 i;
+
+            if (!in.read(i)) {
+                m_error = tr("Unexpected EOF while reading key");
+                return false;
+            }
+
+            out.write(i);
+            break;
+        }
+        default:
+            m_error = tr("Unsupported key part");
+            return false;
+        }
+    }
+
+    return true;
+}
+
 bool OpenSSHKey::readPublic(BinaryStream& stream)
 {
-    m_publicData.clear();
+    // clang-format off
+    static const QMap<QString, QList<KeyPart>> keyTemplates {
+        { "ssh-dss",                            {STR_PART, STR_PART, STR_PART, STR_PART} },
+        { "ssh-rsa",                            {STR_PART, STR_PART} },
+        { "ecdsa-sha2-nistp256",                {STR_PART, STR_PART} },
+        { "ecdsa-sha2-nistp384",                {STR_PART, STR_PART} },
+        { "ecdsa-sha2-nistp521",                {STR_PART, STR_PART} },
+        { "ssh-ed25519",                        {STR_PART} },
+        { "sk-ecdsa-sha2-nistp256@openssh.com", {STR_PART, STR_PART, STR_PART} },
+        { "sk-ssh-ed25519@openssh.com",         {STR_PART, STR_PART} },
+    };
+    // clang-format on
+
+    m_rawPublicData.clear();
+    BinaryStream rawPublicDataStream(&m_rawPublicData);
 
     if (!stream.readString(m_type)) {
         m_error = tr("Unexpected EOF while reading public key");
         return false;
     }
 
-    int keyParts;
-    if (m_type == "ssh-dss") {
-        keyParts = 4;
-    } else if (m_type == "ssh-rsa") {
-        keyParts = 2;
-    } else if (m_type.startsWith("ecdsa-sha2-")) {
-        keyParts = 2;
-    } else if (m_type == "ssh-ed25519") {
-        keyParts = 1;
-    } else {
+    if (!keyTemplates.contains(m_type)) {
         m_error = tr("Unknown key type: %1").arg(m_type);
         return false;
     }
 
-    for (int i = 0; i < keyParts; ++i) {
-        QByteArray t;
-
-        if (!stream.readString(t)) {
-            m_error = tr("Unexpected EOF while reading public key");
-            return false;
-        }
-
-        m_publicData.append(t);
-    }
-
-    return true;
+    return readKeyParts(stream, keyTemplates[m_type], rawPublicDataStream);
 }
 
 bool OpenSSHKey::readPrivate(BinaryStream& stream)
 {
-    m_privateData.clear();
+    // clang-format off
+    static const QMap<QString, QList<KeyPart>> keyTemplates {
+        { "ssh-dss",                            {STR_PART, STR_PART, STR_PART, STR_PART, STR_PART} },
+        { "ssh-rsa",                            {STR_PART, STR_PART, STR_PART, STR_PART, STR_PART, STR_PART} },
+        { "ecdsa-sha2-nistp256",                {STR_PART, STR_PART, STR_PART} },
+        { "ecdsa-sha2-nistp384",                {STR_PART, STR_PART, STR_PART} },
+        { "ecdsa-sha2-nistp521",                {STR_PART, STR_PART, STR_PART} },
+        { "ssh-ed25519",                        {STR_PART, STR_PART} },
+        { "sk-ecdsa-sha2-nistp256@openssh.com", {STR_PART, STR_PART, STR_PART, UINT8_PART, STR_PART, STR_PART} },
+        { "sk-ssh-ed25519@openssh.com",         {STR_PART, STR_PART, UINT8_PART, STR_PART, STR_PART} },
+    };
+    // clang-format on
+
+    m_rawPrivateData.clear();
+    BinaryStream rawPrivateDataStream(&m_rawPrivateData);
 
     if (!stream.readString(m_type)) {
         m_error = tr("Unexpected EOF while reading private key");
         return false;
     }
 
-    int keyParts;
-    if (m_type == "ssh-dss") {
-        keyParts = 5;
-    } else if (m_type == "ssh-rsa") {
-        keyParts = 6;
-    } else if (m_type.startsWith("ecdsa-sha2-")) {
-        keyParts = 3;
-    } else if (m_type == "ssh-ed25519") {
-        keyParts = 2;
-    } else {
+    if (!keyTemplates.contains(m_type)) {
         m_error = tr("Unknown key type: %1").arg(m_type);
         return false;
     }
 
-    for (int i = 0; i < keyParts; ++i) {
-        QByteArray t;
-
-        if (!stream.readString(t)) {
-            m_error = tr("Unexpected EOF while reading private key");
-            return false;
-        }
-
-        m_privateData.append(t);
+    if (!readKeyParts(stream, keyTemplates[m_type], rawPrivateDataStream)) {
+        m_error = tr("Unexpected EOF while reading private key");
+        return false;
     }
 
     if (!stream.readString(m_comment)) {
@@ -525,7 +537,7 @@ bool OpenSSHKey::readPrivate(BinaryStream& stream)
 
 bool OpenSSHKey::writePublic(BinaryStream& stream)
 {
-    if (m_publicData.isEmpty()) {
+    if (m_rawPublicData.isEmpty()) {
         m_error = tr("Can't write public key as it is empty");
         return false;
     }
@@ -535,11 +547,9 @@ bool OpenSSHKey::writePublic(BinaryStream& stream)
         return false;
     }
 
-    for (QByteArray t : m_publicData) {
-        if (!stream.writeString(t)) {
-            m_error = tr("Unexpected EOF when writing public key");
-            return false;
-        }
+    if (!stream.write(m_rawPublicData)) {
+        m_error = tr("Unexpected EOF when writing public key");
+        return false;
     }
 
     return true;
@@ -547,7 +557,7 @@ bool OpenSSHKey::writePublic(BinaryStream& stream)
 
 bool OpenSSHKey::writePrivate(BinaryStream& stream)
 {
-    if (m_privateData.isEmpty()) {
+    if (m_rawPrivateData.isEmpty()) {
         m_error = tr("Can't write private key as it is empty");
         return false;
     }
@@ -557,11 +567,9 @@ bool OpenSSHKey::writePrivate(BinaryStream& stream)
         return false;
     }
 
-    for (QByteArray t : m_privateData) {
-        if (!stream.writeString(t)) {
-            m_error = tr("Unexpected EOF when writing private key");
-            return false;
-        }
+    if (!stream.write(m_rawPrivateData)) {
+        m_error = tr("Unexpected EOF when writing private key");
+        return false;
     }
 
     if (!stream.writeString(m_comment)) {
