@@ -17,12 +17,13 @@
  */
 
 #include "TestGroup.h"
-#include "TestGlobal.h"
 #include "mock/MockClock.h"
 
-#include <QScopedPointer>
+#include <QSet>
 #include <QSignalSpy>
+#include <QtTestGui>
 
+#include "core/Group.h"
 #include "core/Metadata.h"
 #include "crypto/Crypto.h"
 
@@ -314,36 +315,38 @@ void TestGroup::testCopyCustomIcon()
     QScopedPointer<Database> dbSource(new Database());
 
     QUuid groupIconUuid = QUuid::createUuid();
-    QImage groupIcon(16, 16, QImage::Format_RGB32);
-    groupIcon.setPixel(0, 0, qRgb(255, 0, 0));
-    dbSource->metadata()->addCustomIcon(groupIconUuid, groupIcon);
+    QByteArray groupIcon("group icon");
+    QString groupIconName("group icon");
+    dbSource->metadata()->addCustomIcon(groupIconUuid, groupIcon, groupIconName);
 
     QUuid entryIconUuid = QUuid::createUuid();
-    QImage entryIcon(16, 16, QImage::Format_RGB32);
-    entryIcon.setPixel(0, 0, qRgb(255, 0, 0));
-    dbSource->metadata()->addCustomIcon(entryIconUuid, entryIcon);
+    QByteArray entryIcon("entry icon");
+    QString entryIconName("entry icon");
+    dbSource->metadata()->addCustomIcon(entryIconUuid, entryIcon, entryIconName);
 
-    Group* group = new Group();
+    auto* group = new Group();
     group->setParent(dbSource->rootGroup());
     group->setIcon(groupIconUuid);
-    QCOMPARE(group->icon(), groupIcon);
+    QCOMPARE(group->database()->metadata()->customIcon(groupIconUuid).data, groupIcon);
+    QCOMPARE(group->database()->metadata()->customIcon(groupIconUuid).name, groupIconName);
 
-    Entry* entry = new Entry();
+    auto* entry = new Entry();
     entry->setGroup(dbSource->rootGroup());
     entry->setIcon(entryIconUuid);
-    QCOMPARE(entry->icon(), entryIcon);
+    QCOMPARE(entry->database()->metadata()->customIcon(entryIconUuid).data, entryIcon);
+    QCOMPARE(entry->database()->metadata()->customIcon(entryIconUuid).name, entryIconName);
 
     QScopedPointer<Database> dbTarget(new Database());
 
     group->setParent(dbTarget->rootGroup());
     QVERIFY(dbTarget->metadata()->hasCustomIcon(groupIconUuid));
-    QCOMPARE(dbTarget->metadata()->customIcon(groupIconUuid), groupIcon);
-    QCOMPARE(group->icon(), groupIcon);
+    QCOMPARE(dbTarget->metadata()->customIcon(groupIconUuid).data, groupIcon);
+    QCOMPARE(dbTarget->metadata()->customIcon(groupIconUuid).name, groupIconName);
 
     entry->setGroup(dbTarget->rootGroup());
     QVERIFY(dbTarget->metadata()->hasCustomIcon(entryIconUuid));
-    QCOMPARE(dbTarget->metadata()->customIcon(entryIconUuid), entryIcon);
-    QCOMPARE(entry->icon(), entryIcon);
+    QCOMPARE(dbTarget->metadata()->customIcon(entryIconUuid).data, entryIcon);
+    QCOMPARE(dbTarget->metadata()->customIcon(entryIconUuid).name, entryIconName);
 }
 
 void TestGroup::testClone()
@@ -424,39 +427,36 @@ void TestGroup::testCopyCustomIcons()
     QScopedPointer<Database> dbSource(new Database());
     QScopedPointer<Database> dbTarget(new Database());
 
-    QImage iconImage1(1, 1, QImage::Format_RGB32);
-    iconImage1.setPixel(0, 0, qRgb(1, 2, 3));
-
-    QImage iconImage2(1, 1, QImage::Format_RGB32);
-    iconImage2.setPixel(0, 0, qRgb(4, 5, 6));
+    Metadata::CustomIconData icon1 = {QByteArray("icon 1"), "icon 1", Clock::currentDateTimeUtc()};
+    Metadata::CustomIconData icon2 = {QByteArray("icon 2"), "icon 2", Clock::currentDateTimeUtc()};
 
     QScopedPointer<Group> group1(new Group());
     group1->setParent(dbSource->rootGroup());
     QUuid group1Icon = QUuid::createUuid();
-    dbSource->metadata()->addCustomIcon(group1Icon, iconImage1);
+    dbSource->metadata()->addCustomIcon(group1Icon, icon1);
     group1->setIcon(group1Icon);
 
     QScopedPointer<Group> group2(new Group());
     group2->setParent(group1.data());
     QUuid group2Icon = QUuid::createUuid();
-    dbSource->metadata()->addCustomIcon(group2Icon, iconImage1);
+    dbSource->metadata()->addCustomIcon(group2Icon, icon1);
     group2->setIcon(group2Icon);
 
     QScopedPointer<Entry> entry1(new Entry());
     entry1->setGroup(group2.data());
     QUuid entry1IconOld = QUuid::createUuid();
-    dbSource->metadata()->addCustomIcon(entry1IconOld, iconImage1);
+    dbSource->metadata()->addCustomIcon(entry1IconOld, icon1);
     entry1->setIcon(entry1IconOld);
 
     // add history item
     entry1->beginUpdate();
     QUuid entry1IconNew = QUuid::createUuid();
-    dbSource->metadata()->addCustomIcon(entry1IconNew, iconImage1);
+    dbSource->metadata()->addCustomIcon(entry1IconNew, icon1);
     entry1->setIcon(entry1IconNew);
     entry1->endUpdate();
 
     // test that we don't overwrite icons
-    dbTarget->metadata()->addCustomIcon(group2Icon, iconImage2);
+    dbTarget->metadata()->addCustomIcon(group2Icon, icon1);
 
     dbTarget->metadata()->copyCustomIcons(group1->customIconsRecursive(), dbSource->metadata());
 
@@ -468,8 +468,8 @@ void TestGroup::testCopyCustomIcons()
     QVERIFY(metaTarget->hasCustomIcon(entry1IconOld));
     QVERIFY(metaTarget->hasCustomIcon(entry1IconNew));
 
-    QCOMPARE(metaTarget->customIcon(group1Icon).pixel(0, 0), qRgb(1, 2, 3));
-    QCOMPARE(metaTarget->customIcon(group2Icon).pixel(0, 0), qRgb(4, 5, 6));
+    QCOMPARE(metaTarget->customIcon(group1Icon), icon1);
+    QCOMPARE(metaTarget->customIcon(group2Icon), icon1);
 }
 
 void TestGroup::testFindEntry()
@@ -689,66 +689,6 @@ void TestGroup::testPrint()
     QVERIFY(output.contains(QString("subgroup/entry3\n")));
 }
 
-void TestGroup::testLocate()
-{
-    Database* db = new Database();
-
-    Entry* entry1 = new Entry();
-    entry1->setTitle("entry1");
-    entry1->setGroup(db->rootGroup());
-
-    Entry* entry2 = new Entry();
-    entry2->setTitle("entry2");
-    entry2->setGroup(db->rootGroup());
-
-    Group* group1 = new Group();
-    group1->setName("group1");
-    group1->setParent(db->rootGroup());
-
-    Group* group2 = new Group();
-    group2->setName("group2");
-    group2->setParent(group1);
-
-    Entry* entry3 = new Entry();
-    entry3->setTitle("entry3");
-    entry3->setGroup(group1);
-
-    Entry* entry43 = new Entry();
-    entry43->setTitle("entry43");
-    entry43->setGroup(group1);
-
-    Entry* google = new Entry();
-    google->setTitle("Google");
-    google->setGroup(group2);
-
-    QStringList results = db->rootGroup()->locate("entry");
-    QVERIFY(results.size() == 4);
-    QVERIFY(results.contains("/group1/entry43"));
-
-    results = db->rootGroup()->locate("entry1");
-    QVERIFY(results.size() == 1);
-    QVERIFY(results.contains("/entry1"));
-
-    results = db->rootGroup()->locate("Entry1");
-    QVERIFY(results.size() == 1);
-    QVERIFY(results.contains("/entry1"));
-
-    results = db->rootGroup()->locate("invalid");
-    QVERIFY(results.isEmpty());
-
-    results = db->rootGroup()->locate("google");
-    QVERIFY(results.size() == 1);
-    QVERIFY(results.contains("/group1/group2/Google"));
-
-    results = db->rootGroup()->locate("group1");
-    QVERIFY(results.size() == 3);
-    QVERIFY(results.contains("/group1/entry3"));
-    QVERIFY(results.contains("/group1/entry43"));
-    QVERIFY(results.contains("/group1/group2/Google"));
-
-    delete db;
-}
-
 void TestGroup::testAddEntryWithPath()
 {
     Database* db = new Database();
@@ -841,7 +781,7 @@ void TestGroup::testCopyDataFrom()
     group3->setName("TestGroup3");
     group3->customData()->set("testKey", "value");
 
-    QSignalSpy spyGroupModified(group.data(), SIGNAL(groupModified()));
+    QSignalSpy spyGroupModified(group.data(), SIGNAL(modified()));
     QSignalSpy spyGroupDataChanged(group.data(), SIGNAL(groupDataChanged(Group*)));
 
     group->copyDataFrom(group2.data());
@@ -1134,21 +1074,18 @@ void TestGroup::testApplyGroupIconRecursively()
     // Set an icon per UUID to the subgroup and apply recursively
     // -> all groups and entries except the root group have the same icon
     const QUuid subgroupIconUuid = QUuid::createUuid();
-    QImage subgroupIcon(16, 16, QImage::Format_RGB32);
-    subgroupIcon.setPixel(0, 0, qRgb(255, 0, 0));
+    QByteArray subgroupIcon("subgroup icon");
+
     database.metadata()->addCustomIcon(subgroupIconUuid, subgroupIcon);
     subgroup->setIcon(subgroupIconUuid);
     subgroup->applyGroupIconToChildGroups();
     subgroup->applyGroupIconToChildEntries();
     QVERIFY(database.rootGroup()->iconNumber() == rootIconNumber);
     QCOMPARE(subgroup->iconUuid(), subgroupIconUuid);
-    QCOMPARE(subgroup->icon(), subgroupIcon);
     QCOMPARE(subgroupEntry->iconUuid(), subgroupIconUuid);
-    QCOMPARE(subgroupEntry->icon(), subgroupIcon);
     QCOMPARE(subsubgroup->iconUuid(), subgroupIconUuid);
-    QCOMPARE(subsubgroup->icon(), subgroupIcon);
     QCOMPARE(subsubgroupEntry->iconUuid(), subgroupIconUuid);
-    QCOMPARE(subsubgroupEntry->icon(), subgroupIcon);
+    QCOMPARE(subgroup->database()->metadata()->customIcon(subgroupIconUuid).data, subgroupIcon);
 
     // Reset all icons to root icon
     database.rootGroup()->setIcon(rootIconNumber);
@@ -1209,7 +1146,7 @@ void TestGroup::testUsernamesRecursive()
     QVERIFY(usernames.indexOf("Name2") < usernames.indexOf("Name1"));
 }
 
-void TestGroup::testMove()
+void TestGroup::testMoveUpDown()
 {
     Database database;
     Group* root = database.rootGroup();
@@ -1319,3 +1256,40 @@ void TestGroup::testMove()
     QCOMPARE(root->entries().at(2), entry1);
     QCOMPARE(root->entries().at(3), entry0);
 }
+
+void TestGroup::testPreviousParentGroup()
+{
+    Database db;
+    auto* root = db.rootGroup();
+    root->setUuid(QUuid::createUuid());
+    QVERIFY(!root->uuid().isNull());
+    QVERIFY(!root->previousParentGroup());
+    QVERIFY(root->previousParentGroupUuid().isNull());
+
+    auto* group1 = new Group();
+    group1->setUuid(QUuid::createUuid());
+    group1->setParent(root);
+    QVERIFY(!group1->uuid().isNull());
+    QVERIFY(!group1->previousParentGroup());
+    QVERIFY(group1->previousParentGroupUuid().isNull());
+
+    auto* group2 = new Group();
+    group2->setParent(root);
+    group2->setUuid(QUuid::createUuid());
+    QVERIFY(!group2->uuid().isNull());
+    QVERIFY(!group2->previousParentGroup());
+    QVERIFY(group2->previousParentGroupUuid().isNull());
+
+    group1->setParent(group2);
+    QVERIFY(group1->previousParentGroupUuid() == root->uuid());
+    QVERIFY(group1->previousParentGroup() == root);
+
+    // Previous parent shouldn't be recorded if new and old parent are the same
+    group1->setParent(group2);
+    QVERIFY(group1->previousParentGroupUuid() == root->uuid());
+    QVERIFY(group1->previousParentGroup() == root);
+
+    group1->setParent(root);
+    QVERIFY(group1->previousParentGroupUuid() == group2->uuid());
+    QVERIFY(group1->previousParentGroup() == group2);
+}
